"""
Contains the WASFinding class, representing a vulnerability on a web application.
"""

from dataclasses import dataclass, asdict, field
from typing import Dict, Literal
from datetime import datetime
from urllib.parse import unquote_plus
from base64 import b64decode

from ...base.base_list import BaseList


@dataclass
class PayloadRequest:
    """
    Represents a payload request from within a
    WASPayload object.
    """

    method: str = None
    link: str = None
    headers: str = None
    body: str = None

    def __post_init__(self):
        if self.headers:
            try:
                self.headers = b64decode(self.headers).decode("utf-8")
            except UnicodeDecodeError:
                self.headers = self.headers

        if self.body:
            try:
                self.body = b64decode(self.body).decode("utf-8")
            except UnicodeDecodeError:
                self.body = self.body

    def to_dict(self) -> Dict:
        """
        Returns the PayloadRequest as a dictionary.
        """
        return asdict(self)

    def __str__(self) -> str:
        return f"{self.method}={self.link}"

    def __dict__(self) -> Dict:
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a PayloadRequest object from a dictionary.
        """
        return cls(**data)


@dataclass
class PayloadResponce:
    offset: int = 0
    length: int = 0

    def __post_init__(self):
        if not isinstance(self.offset, int):
            self.offset = int(self.offset)
        if not isinstance(self.length, int):
            self.length = int(self.length)

    def to_dict(self) -> Dict:
        """
        Returns the PayloadResponce as a dictionary.
        """
        return asdict(self)

    def __dict__(self) -> Dict:
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a PayloadResponce object from a dictionary.
        """
        return cls(**data)


@dataclass
class WASPayload:
    """
    Represents a payload from within a
    FindingItem object.
    """

    payload: str = None
    request: PayloadRequest = None
    response: str = None
    payloadResponce: PayloadResponce = None

    def __post_init__(self):
        # if self.headers:
        #    self.headers = b64decode(self.headers).decode("utf-8")

        if self.request:
            self.request = PayloadRequest.from_dict(self.request)

        if self.payloadResponce:
            self.payloadResponce = PayloadResponce.from_dict(self.payloadResponce)

    def to_dict(self) -> Dict:
        """
        Returns the WASPayload as a dictionary.
        """
        return asdict(self)

    def __str__(self) -> str:
        return f"{self.payload}"

    def __dict__(self) -> Dict:
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a WASPayload object from a dictionary.
        """
        return cls(**data)


@dataclass
class FindingItem:
    """
    Represents a finding from within a
    WASFinding object.
    """

    authentication: bool = False
    accessPath: dict = None
    # accessPath is parsed into below:
    accessPath_count: int = None
    accessPath_list: BaseList[str] = None
    # end of accessPath
    ajax: bool = False
    ajaxRequestId: str = None
    payloads: dict = None
    # payloads is parsed into below:
    payloads_count: int = None
    payloads_list: BaseList[WASPayload] = None
    # end of payloads
    formLocation: str = None

    def __post_init__(self):
        BOOL_FIELDS = ["ajax", "authentication"]

        for field in BOOL_FIELDS:
            if getattr(self, field):
                setattr(self, field, field == "true")

        if self.accessPath:
            self.accessPath_count = int(self.accessPath.get("count"))
            self.accessPath_list = BaseList()
            if self.accessPath_count > 0:
                data = self.accessPath.get("list").get("Url")
                if not isinstance(data, list):
                    data = [data]
                for itm in data:
                    self.accessPath_list.append(itm)
            setattr(self, "accessPath", None)

        if self.payloads:
            self.payloads_count = int(self.payloads.get("count"))
            self.payloads_list = BaseList()
            if self.payloads_count > 0:
                try:
                    data = (
                        self.payloads.get("list").get("PayloadInstance").get("request")
                    )
                    if data:
                        self.payloads_list.append(PayloadRequest.from_dict(data))
                except AttributeError:
                    data = self.payloads.get("list").get("PayloadInstance")
                    if not isinstance(data, list):
                        data = [data]
                    for itm in data:
                        self.payloads_list.append(WASPayload.from_dict(itm))
            setattr(self, "payloads", None)

    def to_dict(self) -> Dict:
        """
        Returns the FindingItem as a dictionary.
        """
        return asdict(self)

    def __dict__(self) -> Dict:
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a FindingItem object from a dictionary.
        """
        return cls(**data)


@dataclass
class ScanData:
    """
    Represents a scan data object from within
    a WASFinding object.
    """

    id: int = None
    reference: str = None
    launchedDate: datetime = None

    def __post_init__(self):
        if self.id:
            setattr(self, "id", int(self.id))
        if self.launchedDate:
            setattr(
                self,
                "launchedDate",
                datetime.strptime(self.launchedDate, "%Y-%m-%dT%H:%M:%SZ"),
            )

    def to_dict(self) -> Dict:
        """
        Returns the ScanData as a dictionary.
        """
        return asdict(self)

    def __int__(self) -> int:
        return self.id

    def __str__(self) -> str:
        return self.reference

    def __dict__(self) -> Dict:
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a ScanData object from a dictionary.
        """
        return cls(**data)


@dataclass
class WASCItem:
    """
    Represents a WASC item from within a
    WASFinding object.
    """

    name: str = None
    url: str = None
    code: str = None

    def __dict__(self) -> Dict:
        return self.to_dict()

    def to_dict(self) -> Dict:
        """
        Returns the WASCItem as a dictionary.
        """
        return asdict(self)

    def __str__(self) -> str:
        return self.name

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a WASCItem object from a dictionary.
        """
        return cls(**data)


@dataclass
class WASFinding:
    """
    Represents a vulnerability on
    a web application in WAS.
    """

    id: int = None
    uniqueId: str = None
    qid: int = None
    detectionScore: int = None
    name: str = None
    type: Literal["VULNERABILITY", "SENSITIVE_CONTENT", "INFORMATION_GATHERED"] = None
    potential: bool = None
    findingType: Literal["QUALYS", "CUSTOM"] = None
    group: str = None
    severity: Literal[1, 2, 3, 4, 5] = None
    url: str = None
    status: Literal["NEW", "ACTIVE", "REOPENED", "FIXED", "PROTECTED"] = None
    firstDetectedDate: datetime = None
    lastDetectedDate: datetime = None
    lastTestedDate: datetime = None
    fixedDate: datetime = None
    timesDetected: int = None
    webApp: dict = None
    # webApp is parsed into below:
    webApp_id: int = None
    webApp_name: str = None
    webApp_url: str = None
    # end of webApp
    isIgnored: bool = None
    ignoredReason: str = None
    ignoredBy: str = None
    ignoredDate: datetime = None
    ignoredComment: str = None
    reactivateDate: datetime = None
    reactivateIn: int = None
    param: str = None
    cwe: dict = None
    # cwe is parsed into below:
    cwe_count: int = None
    cwe_list: BaseList[int] = None
    # end of cwe
    owasp: dict = None
    # owasp is parsed into below:
    owasp_count: int = None
    owasp_list: BaseList[str] = None
    # end of owasp
    resultList: dict = None
    # resultList is parsed into below:
    resultList_count: int = 0
    resultList_list: BaseList[str] = None
    patch: int = None
    # accessPath_count/list is an aggregate of each
    # object in resultList_list. As is
    # payloads_count/list
    accessPath_count: int = 0
    accessPath_list: BaseList[str] = field(default_factory=BaseList)
    payloads_count: int = 0
    payloads_list: BaseList[WASPayload] = field(default_factory=BaseList)
    # end of resultList
    cvssV3: dict = None
    # cvssV3 is parsed into below:
    cvssV3_base: float = None
    cvssV3_impact: float = None
    cvssV3_attackVector: str = None
    severityComment: str = None
    editedSeverityUser: str = None
    # end of cvssV3
    history: dict = None
    # history is parsed into below:
    history_list: BaseList[ScanData] = None
    # end of history
    wasc: dict = None
    # wasc is parsed into below:
    wasc_count: int = None
    wasc_list: BaseList[WASCItem] = None
    sslData: dict = None
    # sslData is parsed into below:
    sslData_protocol: str = None
    sslData_virtualhost: str = None
    sslData_ip: str = None
    sslData_port: int = None
    sslData_result: str = None
    sslData_list: BaseList[str] = None
    sslData_certificateFingerprint: BaseList[str] = None
    sslData_flags: str = None
    updatedDate: datetime = None

    def __post_init__(self):
        DT_FIELDS = [
            "firstDetectedDate",
            "lastDetectedDate",
            "lastTestedDate",
            "fixedDate",
            "updatedDate",
            "ignoredDate",
            "reactivateDate",
        ]
        INT_FIELDS = [
            "id",
            "qid",
            "detectionScore",
            "severity",
            "timesDetected",
            "webApp_id",
            "patch",
            "reactivateIn",
        ]
        BOOL_FIELDS = ["potential", "isIgnored"]

        if self.webApp:
            self.webApp_id = self.webApp.get("id")
            self.webApp_name = self.webApp.get("name")
            self.webApp_url = self.webApp.get("url")
            setattr(self, "webApp", None)

        if self.status:
            setattr(self, "status", self.status.upper())

        if self.type:
            setattr(self, "type", self.type.upper())

        if self.ignoredBy:
            setattr(self, "ignoredBy", self.ignoredBy.get("id"))

        if self.editedSeverityUser:
            setattr(self, "editedSeverityUser", self.editedSeverityUser.get("id"))

        if self.findingType:
            setattr(self, "findingType", self.findingType.upper())

        if self.param:
            setattr(self, "param", unquote_plus(self.param))

        for field in INT_FIELDS:
            if getattr(self, field):
                setattr(self, field, int(getattr(self, field)))

        for field in BOOL_FIELDS:
            if getattr(self, field):
                setattr(self, field, bool(getattr(self, field)))

        for field in DT_FIELDS:
            if getattr(self, field):
                setattr(
                    self,
                    field,
                    datetime.strptime(getattr(self, field), "%Y-%m-%dT%H:%M:%SZ"),
                )

        if self.cwe:
            setattr(self, "cwe_count", int(self.cwe.get("count")))
            bl = BaseList()
            data = self.cwe.get("list").get("long")
            if data:
                if not isinstance(data, list):
                    data = [data]
                for itm in data:
                    bl.append(int(itm))
            setattr(self, "cwe_list", bl)
            setattr(self, "cwe", None)

        if self.owasp:
            setattr(self, "owasp_count", int(self.owasp.get("count")))
            bl = BaseList()
            data = self.owasp.get("list").get("OWASP")
            if data:
                if not isinstance(data, list):
                    data = [data]
                for itm in data:
                    owaspstr = (
                        f"{itm.get('name')} - {itm.get('url')} - code={itm.get('code')}"
                    )
                    bl.append(owaspstr)
            setattr(self, "owasp_list", bl)
            setattr(self, "owasp", None)

        if self.resultList:
            setattr(self, "resultList_count", int(self.resultList.get("count")))
            bl = BaseList()
            data = self.resultList.get("list").get("Result")
            if data:
                if not isinstance(data, list):
                    data = [data]
                for itm in data:
                    bl.append(FindingItem.from_dict(itm))

            if bl:
                # Parse out accessPath_count/list, payloads_count/list
                for itm in bl:
                    if itm.accessPath_list:
                        self.accessPath_count += itm.accessPath_count
                        self.accessPath_list += itm.accessPath_list
                    if itm.payloads_list:
                        self.payloads_count += itm.payloads_count
                        self.payloads_list += itm.payloads_list

            setattr(self, "resultList_list", bl)
            setattr(self, "resultList", None)

        if self.cvssV3:
            self.cvssV3_base = float(self.cvssV3.get("base"))
            self.cvssV3_temporal = float(self.cvssV3.get("temporal"))
            self.cvssV3_attackVector = self.cvssV3.get("attackVector")
            setattr(self, "cvssV3", None)

        if self.history:
            data = self.history["set"]["WebAppFindingHistory"]
            bl = BaseList()
            if data:
                if not isinstance(data, list):
                    data = [data]
                for itm in data:
                    bl.append(ScanData.from_dict(itm["scanData"]))
            setattr(self, "history_list", bl)
            setattr(self, "history", None)

        if self.wasc:
            setattr(self, "wasc_count", int(self.wasc.get("count")))
            bl = BaseList()
            data = self.wasc.get("list").get("WASC")
            if data:
                if not isinstance(data, list):
                    data = [data]
                for itm in data:
                    bl.append(WASCItem.from_dict(itm))
            setattr(self, "wasc_list", bl)
            setattr(self, "wasc", None)

        if self.sslData:
            self.sslData_certificateFingerprint = BaseList()
            self.sslData_protocol = self.sslData.get("protocol")
            self.sslData_virtualhost = self.sslData.get("virtualhost")
            self.sslData_ip = self.sslData.get("ip")
            self.sslData_flags = self.sslData.get("flags")
            try:
                self.sslData_port = int(self.sslData.get("port"))
            except TypeError:
                self.sslData_port = None

            self.sslData_result = self.sslData.get("result")
            data = self.sslData.get("sslDataInfoList")
            if data:
                data = data["list"].get("SSLDataInfo")
                bl = BaseList()
                if not isinstance(data, list):
                    data = [data]
                for datapoint in data:
                    for itm, val in datapoint.items():
                        match itm:
                            case "sslDataCipherList":
                                if val:
                                    for cipher in val.get("list").get("SSLDataCipher"):
                                        # protocol, name, keyExchange, auth, mac, encryption, grade
                                        string = f"{cipher.get('protocol')}: {cipher.get('name')} (keyExchange: {cipher.get('keyExchange')}; auth: {cipher.get('auth')}; mac: {cipher.get('mac')}; encryption: {cipher.get('encryption')}; grade: {cipher.get('grade')})"
                                        bl.append(string) if string not in bl else None
                            case "sslDataKexList":
                                if val:
                                    for kex in val.get("list").get("SSLDataKex"):
                                        string = f"{kex.get('protocol')}: {kex.get('kex')} (keysize: {kex.get('keysize')}; fwdsec: {kex.get('fwdsec')}; classical: {kex.get('classical')}; quantum: {kex.get('quantum')})"
                                        bl.append(string) if string not in bl else None
                            case "sslDataPropList":
                                if val:
                                    for prop in val.get("list").get("SSLDataProp"):
                                        string = f"{prop.get('name')}: {prop.get('value')} ({prop.get('protocol')})"
                                        bl.append(string) if string not in bl else None
                            case "certificateFingerprint":
                                self.sslData_certificateFingerprint.append(val)

                self.sslData_list = bl
            setattr(self, "sslData", None)

    def to_dict(self) -> Dict:
        """
        Returns the WASFinding as a dictionary.
        """
        return asdict(self)

    def __dict__(self) -> Dict:
        return self.to_dict()

    def __int__(self) -> int:
        return self.id

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    def __getitem__(self, key):
        return self.to_dict().get(key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __eq__(self, other):
        if not isinstance(other, WASFinding):
            return False
        return self.id == other.id

    @classmethod
    def from_dict(cls, data: Dict):
        """
        Returns a WASFinding object from a dictionary.
        """
        return cls(**data)



"""
Contains the CrawlingScript class for WAS
"""

from dataclasses import dataclass, asdict
from typing import Union


@dataclass
class CrawlingScript:
    """
    Represents a crawl scriptin WAS
    """

    id: int = None
    name: str = None
    data: str = None
    regex: str = None
    requiresAuthentication: Union[str, bool] = None
    startingUrl: str = None
    startingUrlRegex: Union[str, bool] = None

    def __post_init__(self):
        if self.id and isinstance(self.id, str):
            try:
                self.id = int(self.id)
            except ValueError:
                raise ValueError("id must be numeric")

        if self.requiresAuthentication and isinstance(self.requiresAuthentication, str):
            match self.requiresAuthentication.lower():
                case "true":
                    self.requiresAuthentication = True
                case "false":
                    self.requiresAuthentication = False
                case _:
                    raise ValueError("requiresAuthentication must be a boolean")

        if self.startingUrlRegex and isinstance(self.startingUrlRegex, str):
            match self.startingUrlRegex.lower():
                case "true":
                    self.startingUrlRegex = True
                case "false":
                    self.startingUrlRegex = False
                case _:
                    raise ValueError("startingUrlRegex must be a boolean")

    def __str__(self):
        return self.name

    def __int__(self):
        return self.id

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @staticmethod
    def from_dict(data):
        return CrawlingScript(**data)



"""
Basic functionality / helpers for qualysdk.
"""

from .call_api import call_api
from .call_schema import CALL_SCHEMA
from .xml_parser import xml_parser
from .base_list import BaseList
from .csv_export import write_csv, write_excel



"""
kb_entry.py - contains the KBEntry class for the qualysdk package.

This class is used to represent a single entry in the Qualys KnowledgeBase (KB).
"""

from dataclasses import dataclass, field, asdict
from typing import *
from datetime import datetime
from warnings import catch_warnings, simplefilter

from bs4 import BeautifulSoup

from ...base.base_list import BaseList
from .bugtraq import Bugtraq
from .software import Software
from .vendor_reference import VendorReference
from .cve import CVEID
from .threat_intel import ThreatIntel
from .compliance import Compliance


@dataclass(order=True)
class KBEntry:
    """
    KBEntry - represents a single entry in the Qualys KnowledgeBase (KB).

    Made with order=True to allow sorting.
    """

    QID: Union[str, int] = field(
        compare=True,
        metadata={"description": "The Qualys ID of the vulnerability."},
        default=None,
    )
    VULN_TYPE: str = field(
        metadata={"description": "The type of vulnerability."}, default="Vulnerability"
    )
    SEVERITY_LEVEL: int = field(
        metadata={"description": "The severity level of the vulnerability."},
        default=None,
    )
    TITLE: str = field(
        metadata={"description": "The title of the vulnerability."}, default=None
    )
    CATEGORY: str = field(
        metadata={"description": "The category of the vulnerability."},
        default=None,
    )
    LAST_SERVICE_MODIFICATION_DATETIME: Optional[datetime] = field(
        metadata={
            "description": "The date the vulnerability was last modified by the service."
        },
        default=None,
    )
    PUBLISHED_DATETIME: Optional[datetime] = field(
        metadata={"description": "The date the vulnerability was published."},
        default=None,
    )
    CODE_MODIFIED_DATETIME: Optional[datetime] = field(
        metadata={"description": "The date the vulnerability code was last modified."},
        default=None,
    )
    BUGTRAQ_LIST: Optional[BaseList] = field(
        metadata={
            "description": "A list of Bugtraq IDs affected by the vulnerability."
        },
        default=None,
    )
    PATCHABLE: Optional[bool] = field(
        metadata={"description": "Whether the vulnerability is patchable."},
        default=False,
    )
    SOFTWARE_LIST: Optional[BaseList] = field(
        metadata={"description": "A list of software affected by the vulnerability."},
        default=None,
    )
    VENDOR_REFERENCE_LIST: Optional[BaseList] = field(
        metadata={
            "description": "A list of vendor bulletin references for the vulnerability."
        },
        default=None,
    )
    CVE_LIST: Optional[BaseList] = field(
        metadata={"description": "A list of CVEIDs affected by the vulnerability."},
        default=None,
    )
    DIAGNOSIS: Optional[str] = field(
        metadata={"description": "The diagnosis of the vulnerability."}, default=""
    )
    CONSEQUENCE: Optional[str] = field(
        metadata={"description": "The consequence of the vulnerability."}, default=None
    )
    SOLUTION: Optional[str] = field(
        metadata={"description": "The solution to the vulnerability."}, default=None
    )
    CORRELATION: Optional[dict] = field(
        metadata={"description": "The correlation details of the vulnerability."},
        default=None,
    )  # TODO... maybe...
    CVSS: Optional[dict] = field(
        metadata={"description": "The CVSS score of the vulnerability."}, default=None
    )
    CVSS_V3: Optional[dict] = field(
        metadata={"description": "The CVSS v3 score of the vulnerability."},
        default=None,
    )
    PCI_FLAG: Optional[bool] = field(
        metadata={"description": "Whether the vulnerability is a PCI flag."},
        default=False,
    )
    PCI_REASONS: Optional[dict] = field(
        metadata={
            "description": "The reasons the vulnerability is valid for PCI requirements."
        },
        default=None,
    )
    THREAT_INTELLIGENCE: Optional[BaseList] = field(
        metadata={
            "description": "The threat intelligence details of the vulnerability."
        },
        default=None,
    )
    SUPPORTED_MODULES: Optional[str] = field(
        metadata={"description": "The supported modules for the vulnerability."},
        default=None,
    )
    DISCOVERY: Optional[dict] = field(
        metadata={"description": "The discovery details of the vulnerability."},
        default=None,
    )
    IS_DISABLED: Optional[bool] = field(
        metadata={"description": "Whether the vulnerability is disabled."},
        default=False,
    )
    CHANGE_LOG: Optional[dict] = field(
        metadata={"description": "The change log of the vulnerability."}, default=None
    )
    TECHNOLOGY: Optional[str] = field(
        metadata={"description": "The technology of the vulnerability."},
        default=None,
    )
    COMPLIANCE_LIST: Optional[BaseList] = field(
        metadata={
            "description": "The list of compliance frameworks for the vulnerability."
        },
        default=None,
    )
    LAST_CUSTOMIZATION: Optional[datetime] = field(
        metadata={"description": "The date the vulnerability was last customized."},
        default=None,
    )
    SOLUTION_COMMENT: Optional[str] = field(
        metadata={"description": "The solution comment for the vulnerability."},
        default=None,
    )

    def __post_init__(self):
        # convert certain fields out of string format:
        if self.QID is not None and not isinstance(self.QID, int):
            self.QID = int(self.QID)

        if self.SEVERITY_LEVEL and not isinstance(self.SEVERITY_LEVEL, int):
            self.SEVERITY_LEVEL = int(self.SEVERITY_LEVEL)

        DATE_FIELDS = [
            "LAST_SERVICE_MODIFICATION_DATETIME",
            "PUBLISHED_DATETIME",
            "CODE_MODIFIED_DATETIME",
            "LAST_CUSTOMIZATION",
        ]

        BOOL_FIELDS = ["PATCHABLE", "PCI_FLAG", "IS_DISABLED"]

        HTML_FIELDS = ["DIAGNOSIS", "CONSEQUENCE", "SOLUTION"]

        for field in DATE_FIELDS:
            if getattr(self, field) is not None and not isinstance(
                getattr(self, field), datetime
            ):
                # special case for LAST_CUSTOMIZATION:
                if field == "LAST_CUSTOMIZATION":
                    if isinstance(getattr(self, field), dict):
                        setattr(
                            self,
                            field,
                            datetime.fromisoformat(getattr(self, field)["DATETIME"]),
                        )
                    else:
                        setattr(
                            self, field, datetime.fromisoformat(getattr(self, field))
                        )
                else:
                    setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        for field in BOOL_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), bool):
                setattr(self, field, bool(getattr(self, field)))

        with catch_warnings():
            simplefilter("ignore")  # ignore the warning about the html.parser
            for field in HTML_FIELDS:
                if getattr(self, field):
                    setattr(
                        self,
                        field,
                        BeautifulSoup(getattr(self, field), "html.parser").get_text(),
                    )

        # convert the lists to BaseList objects:
        if self.BUGTRAQ_LIST:
            final_bugtraq_list = BaseList()
            data = self.BUGTRAQ_LIST["BUGTRAQ"]
            if isinstance(data, dict):
                # Put into a list for easier processing:
                data = [data]

            for bugtraq in data:
                final_bugtraq_list.append(Bugtraq.from_dict(bugtraq))

            self.BUGTRAQ_LIST = final_bugtraq_list

        if self.SOFTWARE_LIST:
            final_software_list = BaseList()
            data = self.SOFTWARE_LIST["SOFTWARE"]
            if isinstance(data, dict):
                # Put into a list for easier processing:
                data = [data]

            for sw in data:
                final_software_list.append(Software.from_dict(sw))

            self.SOFTWARE_LIST = final_software_list

        if self.VENDOR_REFERENCE_LIST:
            final_vendor_reference_list = BaseList()
            data = self.VENDOR_REFERENCE_LIST["VENDOR_REFERENCE"]
            if isinstance(data, dict):
                # Put into a list for easier processing:
                data = [data]

            for vendor_ref in data:
                final_vendor_reference_list.append(
                    VendorReference.from_dict(vendor_ref)
                )

            self.VENDOR_REFERENCE_LIST = final_vendor_reference_list

        if self.CVE_LIST:
            final_cve_list = BaseList()
            data = self.CVE_LIST["CVE"]
            if isinstance(data, dict):
                # Put into a list for easier processing:
                data = [data]

            for cve in data:
                final_cve_list.append(CVEID.from_dict(cve))

            self.CVE_LIST = final_cve_list

        if self.THREAT_INTELLIGENCE:
            final_threat_intel_list = BaseList()
            data = self.THREAT_INTELLIGENCE["THREAT_INTEL"]
            if isinstance(data, dict):
                # Put into a list for easier processing:
                data = [data]

            for threat_intel in data:
                # Ensure @id-> ID and #text-> TEXT
                threat_intel["ID"] = threat_intel.pop("@id")
                threat_intel["TEXT"] = threat_intel.pop("#text")

                final_threat_intel_list.append(ThreatIntel.from_dict(threat_intel))

            self.THREAT_INTELLIGENCE = final_threat_intel_list

        if self.COMPLIANCE_LIST:
            final_compliance_list = BaseList()
            data = self.COMPLIANCE_LIST["COMPLIANCE"]
            if isinstance(data, dict):
                # Put into a list for easier processing:
                data = [data]

            for compliance in data:
                # Ensure TYPE -> _TYPE
                compliance["_TYPE"] = compliance.pop("TYPE")
                final_compliance_list.append(Compliance.from_dict(compliance))

            self.COMPLIANCE_LIST = final_compliance_list

    def __str__(self):
        return f"{self.QID}"

    def __eq__(self, other):
        return self.QID == other.QID

    def __hash__(self):
        return hash(self.QID)

    def __iter__(self):
        for key, value in self.items():
            yield key, value

    def __contains__(self, item):
        return item in self.QID or item in self.TITLE

    def copy(self):
        return KBEntry.from_dict(self.to_dict())

    def is_qid(self, qid: int):
        return self.QID == qid

    def items(self):
        return self.to_dict().items()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        # return the values
        return self.to_dict().values()

    def __dict__(self):
        return asdict(self)

    def to_dict(self):
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a KBEntry object from a dictionary.

        This function is used to create a KBEntry object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"QID", "VULN_TYPE", "SEVERITY_LEVEL", "TITLE", "CATEGORY"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        # convert the datetime strings to datetime objects:
        for key in [
            "PUBLISHED_DATETIME",
            "CODE_MODIFIED_DATETIME",
            "LAST_SERVICE_MODIFICATION_DATETIME",
        ]:
            if key in data:
                data[key] = datetime.fromisoformat(data[key])

        # and finally, create the KBEntry object:
        return cls(**data)



"""
get_asset.py - get a specific asset from the Global AssetView API via its asset ID.
"""

from ..base.call_api import call_api
from ..auth.token import TokenAuth
from ..exceptions.Exceptions import *
from .hosts import Host


def get_asset(auth: TokenAuth, **kwargs) -> Host:
    """
    Get a specific host from the Global AssetView API.

    Params:
        auth (TokenAuth): The authentication object.

    :Kwargs:
        assetId (int): The asset ID to get.
        lastSeenAssetId (int): The last seen asset ID.
        lastModifiedDate (str): The last modified date.

    Returns:
        Host: The Host object.
    """
    # despite the fact that this is a POST request, we still need to send stuff as a parameter
    # because Qualys is Qualys.

    # make the request:
    response = call_api(auth=auth, module="gav", endpoint="get_asset", params=kwargs)

    # check for 204 response:
    if response.status_code == 204:
        raise ValueError(f"No asset found with ID {kwargs['assetId']}.")

    # parse the response:
    j = response.json()

    if response.status_code != 200:
        raise QualysAPIError(j)

    return Host(**j["assetListData"]["asset"][0])



"""
Contains the UrlEntry class for WAS
"""

from dataclasses import dataclass, asdict
from typing import Union


@dataclass
class UrlEntry:
    """
    Represents a single URL entry in WAS
    """

    regex: Union[str, bool] = None
    text: str = None

    def __post_init__(self):
        if self.regex and isinstance(self.regex, str):
            match self.regex.lower():
                case "true":
                    self.regex = True
                case "false":
                    self.regex = False
                case _:
                    raise ValueError("regex must be a boolean")
        else:
            self.regex = False

    def __str__(self):
        return self.text

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @staticmethod
    def from_dict(data):
        return UrlEntry(**data)



"""
Contains write_csv and write_excel functions,
which wrap pandas.DataFrame.to_csv and pandas.DataFrame.to_excel
to accurately export data to CSV and Excel files.
"""

import csv
from typing import Literal

from pandas import DataFrame

from ..sql.base import prepare_dataclass
from .base_list import BaseList


def write_csv(data: BaseList, file_path: str, **kwargs) -> None:
    """
    Write data to a CSV file.

    Args:
        data (BaseList): The data to write.
        file_path (str): The path to the file to write.

    ## Kwargs:

        - **kwargs: Additional keyword arguments to pass to pandas.DataFrame.to_csv other than quoting, sep, and index.
    """

    data = backend_write(data, file_path, "csv", **kwargs)

    # If user path does not end in .csv, add it:
    if not file_path.lower().endswith(".csv"):
        file_path += ".csv"

    # Write to CSV:
    data.to_csv(file_path, index=False, sep=",", quoting=csv.QUOTE_ALL, **kwargs)
    print(f"Data written to {file_path}.")


def write_excel(data: BaseList, file_path: str, **kwargs) -> None:
    """
    Write data to an Excel file.

    Args:
        data (BaseList): The data to write.
        file_path (str): The path to the file to write.

    ## Kwargs:

        - **kwargs: Additional keyword arguments to pass to pandas.DataFrame.to_excel other than index.
    """

    data = backend_write(data, file_path, "excel", **kwargs)

    # If user path does not end in .xlsx, add it:
    if not file_path.lower().endswith(".xlsx"):
        file_path += ".xlsx"

    # Write to Excel:
    data.to_excel(file_path, index=False, **kwargs)
    print(f"Data written to {file_path}.")


def backend_write(
    data: BaseList, file_path: str, file_type: Literal["csv", "excel"], **kwargs
) -> DataFrame:
    """
    Write data to a file.

    Args:
        data (BaseList): The data to write.
        file_path (str): The path to the file to write.

    ## Kwargs:

        - **kwargs: Additional keyword arguments to pass to pandas.DataFrame.to_csv or pandas.DataFrame.to_excel other than index.
    """

    # Convert dataclass to dictionary if needed.
    # Check that all things in the BaseList have
    # the to_dict method.
    if all(hasattr(thing, "to_dict") for thing in data):
        data = [prepare_dataclass(thing) for thing in data]
    else:
        raise ValueError(
            "All items in the BaseList must have a to_dict method. Bad indexes: ",
            [i for i, thing in enumerate(data) if not hasattr(thing, "to_dict")],
        )

    # Convert to DataFrame if needed:
    if not isinstance(data, DataFrame):
        data = DataFrame(data)

    # Make any datetime cols timezone unaware:
    for col in data.columns:
        if data[col].dtype in [
            "datetime64[ns]",
            "datetime64[ns, tz]",
            "datetime64[ns, UTC]",
            "timedelta64[ns]",
        ]:
            data[col] = data[col].dt.tz_localize(None)

    return data



"""
TotalCloud QID class
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime


@dataclass
class QID:
    """
    Represents a Qualys QID for the TotalCloud module
    """

    qid: int = None
    patchAvailable: bool = None
    exploitability: str = None
    severity: int = None
    customerSeverity: int = None
    lastFound: Union[str, datetime] = None
    firstFound: Union[str, datetime] = None
    port: int = None
    protocol: str = None
    category: str = None
    hostOS: str = None
    typeDetected: str = None
    status: str = None
    disabled: bool = None
    ignored: bool = None

    def __post_init__(self):
        DT_FIELDS = ["lastFound", "firstFound"]
        for field in DT_FIELDS:
            if isinstance(getattr(self, field), str):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if getattr(self, "exploitability") and getattr(self, "exploitability") == "---":
            setattr(self, "exploitability", None)

    def __int__(self):
        return self.qid

    def __dict__(self):
        return asdict(self)

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)



"""
call_api.py - contains the call_api function for the qualysdk package.

This function handles all API calls to the Qualys API. It takes in a URL, headers, and a payload, and returns the response from the API.
Qualys uses many tricks in their API, such as using both url params and post data.
"""

from requests import request, Response
from typing import Literal, Union
from datetime import datetime, timedelta
from time import sleep

from ..auth.token import TokenAuth
from ..auth.basic import BasicAuth
from ..exceptions.Exceptions import *
from .call_schema import CALL_SCHEMA
from .convert_bools_and_nones import convert_bools_and_nones
from .xml_parser import xml_parser


def call_api(
    auth: Union[BasicAuth, TokenAuth],
    module: str,
    endpoint: str,
    headers: dict = None,
    params: dict = None,
    payload: dict = None,
    jsonbody: dict = None,
    override_method: Literal["GET", "POST", "PUT", "PATCH", "DELETE"] = None,
) -> Response:
    """
    Base call function for the Qualys API.

    This function is used across all modules and endpoints as the actual
    API call function.

    Params:
    ```
    auth (Union[BasicAuth, TokenAuth]) The authentication object.
    module (str) The module to call using the CALL_SCHEMA.
    endpoint (str) The endpoint to call using the CALL_SCHEMA.
    headers (dict) The headers to send.
    payload (dict) The payload to send.
    params (dict) The parameters to send.
    jsonbody (dict) The JSON body to send.
    override_method (Literal["GET", "POST", "PUT", "PATCH", "DELETE"]) The method to override the schema with.
    ```
    """
    while True:  # loop to handle hitting the rate limit
        # check module and endpoint:
        if module.lower() not in CALL_SCHEMA.keys():
            raise ValueError(
                f"Invalid module {module}. Valid modules are: {CALL_SCHEMA.keys()}."
            )
        if endpoint.lower() not in CALL_SCHEMA[module].keys():
            raise ValueError(
                f"Invalid endpoint {endpoint} for module {module}. Valid endpoints are: {[i for i in CALL_SCHEMA[module].keys() if i != 'url_type']}."
            )

        SCHEMA = CALL_SCHEMA[module][endpoint]

        # check the auth type:
        if SCHEMA["auth_type"] != auth.auth_type:
            raise AuthTypeMismatchError(
                f"Auth type mismatch. Expected {SCHEMA['auth_type']} but got {auth.auth_type}."
            )

        # check override:
        if override_method:
            if override_method.upper() not in SCHEMA["method"]:
                raise ValueError(
                    f"Invalid override method {override_method}. Valid methods are: {SCHEMA['method']}."
                )

        # match the url_type to get the proper template:
        match CALL_SCHEMA[module]["url_type"]:
            case "gateway":
                url = f"https://gateway.{auth.platform}.apps.qualys.com{SCHEMA['endpoint']}"
            case "api":
                if (
                    auth.platform == "qg1"
                ):  # Special case for qg1 platform: no qg or apps in the URL
                    url = f"https://qualysapi.qualys.com{SCHEMA['endpoint']}"
                else:
                    url = f"https://qualysapi.{auth.platform}.apps.qualys.com{SCHEMA['endpoint']}"
            case "base":
                if auth.platform == "qg1":
                    url = f"https://qualysguard.qualys.com{SCHEMA['endpoint']}"
                else:
                    url = f"https://qualysguard.{auth.platform}.apps.qualys.com{SCHEMA['endpoint']}"
            case _:
                raise ValueError(f"Invalid url_type {SCHEMA['url_type']}.")

        # if token auth, check if token is not 4+ hours old:
        if isinstance(auth, TokenAuth):
            # check that the time delta between now and the token generation time is less than ~4 hours:
            if (datetime.now() - auth.generated_on).seconds > 14395:
                print("Token is 4+ hours old. Refreshing token...")
                auth.get_token()

        # check params:
        if params:
            for key in params.keys():
                if key not in SCHEMA["valid_params"]:
                    raise ValueError(
                        f"Invalid parameter {key} for {module}-{endpoint}. Valid parameters are: {SCHEMA['valid_params'].sort() if SCHEMA['valid_params'].sort() else SCHEMA['valid_params']}."
                    )

        # check post data:
        if payload:
            for key in payload.keys():
                if key not in SCHEMA["valid_POST_data"]:
                    raise ValueError(
                        f"Invalid payload key {key} for {module}-{endpoint}."
                    )

        # check if data should be POSTed as requests.post(json=):
        if SCHEMA["use_requests_json_data"]:
            use_json = True
        else:
            use_json = False

        # set up JWT auth header if needed:
        if auth.auth_type == "token":
            headers = auth.as_header()
        # or set up the tuple for basic auth:
        elif auth.auth_type == "basic":
            auth_tuple = (auth.username, auth.password)

        # Make certain payloads/params requests-friendly:
        if payload:
            payload = convert_bools_and_nones(payload)
        if params:
            params = convert_bools_and_nones(params)

        # If the URL of an endpoint has the substring
        # {placeholder}, .pop() from the params/payload
        # and format the url with the value:
        if any(
            keyword in url
            for keyword in [
                "{placeholder}",
                "{cloudprovider}",
                "{connectorid}",
                "{controlid}",
                "{webappId}",
                "{findingId}",
            ]
        ):
            if params and (
                any(
                    keyword in params
                    for keyword in [
                        "placeholder",
                        "cloudprovider",
                        "connectorid",
                        "controlid",
                        "webappId",
                        "findingId",
                    ]
                )
            ):
                url = url.format(
                    placeholder=str(params.pop("placeholder", None)),
                    cloudprovider=str(params.pop("cloudprovider", None)),
                    connectorid=str(params.pop("connectorid", None)),
                    controlid=str(params.pop("controlid", None)),
                    resourceid=str(params.pop("resourceid", None)),
                    webappId=str(params.pop("webappId", None)),
                    webappAuthRecordId=str(params.pop("webappAuthRecordId", None)),
                    findingId=str(params.pop("findingId", None)),
                )
            elif payload and (
                any(
                    keyword in payload
                    for keyword in [
                        "placeholder",
                        "cloudprovider",
                        "connectorid",
                        "controlid",
                        "resourceid",
                        "webappId",
                        "findingId",
                    ]
                )
            ):
                url = url.format(
                    placeholder=str(payload.pop("placeholder", None)),
                    cloudprovider=str(payload.pop("cloudprovider", None)),
                    connectorid=str(payload.pop("connectorid", None)),
                    controlid=str(payload.pop("controlid", None)),
                    resourceid=str(payload.pop("resourceid", None)),
                    webappId=str(payload.pop("webappId", None)),
                    findingId=str(payload.pop("findingId", None)),
                )
            else:
                raise ValueError(
                    f"Endpoint {module}-{endpoint} requires a placeholder or cloudprovider value in the URL however none was found in params/POST data. Base URL is: {url}"
                )

        # If _xml_data key is defined in call schema,
        # use it as the payload/params:
        if payload and SCHEMA.get("_xml_data") and payload.get("_xml_data"):
            payload = payload["_xml_data"]

        if params and SCHEMA.get("_xml_data") and params.get("_xml_data"):
            params = params["_xml_data"]

        # and finally, make the request:
        response = request(
            method=(
                SCHEMA["method"][0] if not override_method else override_method.upper()
            ),
            url=url,
            headers=headers,
            params=params,
            data=payload if not use_json else None,
            json=jsonbody if use_json else None,
            auth=(auth_tuple if auth.auth_type == "basic" else None),
        )

        # check for errors not related to rate limiting:
        if (
            module
            not in ["gav", "cloud_agent", "cloudview", "containersecurity", "was"]
            and response.status_code in range(400, 599)
            and response.status_code not in [429, 414]
            and (
                response.status_code != 409
                and "This API cannot be run again for another"
                not in xml_parser(response.text)
                .get("SIMPLE_RETURN")
                .get("RESPONSE")
                .get("TEXT")
            )
        ):
            parsed = xml_parser(response.text) if module not in ["gav"] else None

            # Common path is [SIMPLE_RETURN][RESPONSE][TEXT] for XML
            if parsed:
                if "SIMPLE_RETURN" in parsed:
                    raise QualysAPIError(parsed["SIMPLE_RETURN"]["RESPONSE"]["TEXT"])
                elif "html" in parsed:
                    raise Exception(
                        f"Error: {parsed['html']['body']['h1']}: {parsed['html']['body']['p'][1]['#text']}"
                    )
                else:
                    # yeah... turns out a homegrown xml parsers aren't super easy...
                    raise Exception(
                        f"Error: {parsed['{http://www.w3.org/1999/xhtml}html']['{http://www.w3.org/1999/xhtml}body']['{http://www.w3.org/1999/xhtml}h1']}"
                    )
            else:  # or JSON
                raise QualysAPIError(
                    response.text if response.text else response.reason
                )

        # Check rate limit details from headers.
        if (
            "X-RateLimit-Remaining" in response.headers
            and int(response.headers["X-RateLimit-Remaining"]) < 10
        ) or response.status_code == 429:
            # Rate limit reached:
            if int(response.headers["X-RateLimit-Remaining"]) == 0:
                # Call API again for the X-RateLimit-ToWait-Sec header.
                # Qualys sometimes only includes this header when the rate limit is reached and retried:
                response = request(
                    method=(
                        SCHEMA["method"][0]
                        if not override_method
                        else override_method.upper()
                    ),
                    url=url,
                    headers=headers,
                    params=params,
                    data=payload if not use_json else None,
                    json=jsonbody if use_json else None,
                    auth=(auth_tuple if auth.auth_type == "basic" else None),
                )

                # Isolate the wait time from the header:
                to_wait = response.headers.get("X-RateLimit-ToWait-Sec")
                if to_wait:
                    to_wait = (
                        int(to_wait) + 3
                    )  # Add 3 seconds to the wait time to be safe.
                else:
                    to_wait = 3601  # Default to 1h 1s if no header is present.

                print(
                    "WARNING: You have reached the rate limit for this endpoint. "
                    + f"qualysdk will automatically sleep for {to_wait} seconds and try again at approximately {datetime.now() + timedelta(seconds=to_wait)}."
                )
                sleep(to_wait)
                # Go to next iteration of the loop to try again:
                continue

            else:
                # Almost at rate limit:
                print(
                    f"Warning: This endpoint will accept {response.headers['X-RateLimit-Remaining']} more calls before rate limiting you."
                    + f" qualysdk will automatically sleep once remaining calls hits 0."
                )

        return response



"""
Contains the functions to upload supported Container Security API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_was_webapps(
    webapps: BaseList,
    cnxn: Connection,
    table_name: str = "was_webapps",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```was.get_webapps```
    or ```was.get_webapps_verbose```
    to a SQL database.

    Args:
        webapps (BaseList): A BaseList of WebApp objects.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to.
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.Integer(),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "url": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "uris": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "riskScore": types.Integer(),
        "os": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "owner_id": types.Integer(),
        "owner_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owner_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owner_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scope": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "attributes": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultProfile_id": types.Integer(),
        "defaultProfile_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultScanner_id": types.Integer(),
        "defaultScanner_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultScannerTags_count": types.Integer(),
        "defaultScannerTags_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scannerLocked": types.Boolean(),
        "progressiveScanning": types.Boolean(),
        "urlExcludelist": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "urlAllowlist": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "postDataExcludelist": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "logoutRegexList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "authRecords": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dnsOverrides": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "useRobots": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "useSitemap": types.Boolean(),
        "malwareMonitoring": types.Boolean(),
        "malwareNotifications": types.Boolean(),
        "tags": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "comments": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isScheduled": types.Boolean(),
        "lastScan_id": types.Integer(),
        "lastScan_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "lastScan_summary_resultsStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "lastScan_summary_authStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_id": types.Integer(),
        "createdBy_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdDate": types.DateTime(),
        "updatedDate": types.DateTime(),
        "updatedBy_id": types.Integer(),
        "updatedBy_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "screenshot": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "config": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "crawlingScripts": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "postmanCollection": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "headers": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "domains": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subDomain": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "swaggerFile": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "redundancyLinks": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "maxRedundancyLinks": types.Integer(),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(webapp) for webapp in webapps])

    # Drop any columns that we parsed out:
    df.drop(
        columns=[
            "owner",
            "defaultProfile",
            "defaultScanner",
            "defaultScannerTags",
            "lastScan",
            "createdBy",
            "updatedBy",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)


def upload_was_authentication_records(
    authRecords: BaseList,
    cnxn: Connection,
    table_name: str = "was_authentication_records",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```was.get_authentication_records```
    or ```was.get_authentication_records_verbose``` to a SQL database.

    Args:
        authRecords (BaseList): A BaseList of WebAppAuthRecord objects.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to.
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.Integer(),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "owner_id": types.Integer(),
        "owner_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owner_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owner_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "formRecord_type": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "formRecord_sslOnly": types.Boolean(),
        "formRecord_authVault": types.Boolean(),
        "formRecord_seleniumCreds": types.Boolean(),
        "formRecord_fields_count": types.Integer(),
        "formRecord_fields_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "serverRecord_type": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "serverRecord_sslOnly": types.Boolean(),
        "serverRecord_authVault": types.Boolean(),
        "serverRecord_fields_count": types.Integer(),
        "serverRecord_fields_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "oauth2Record_grantType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "oauth2Record_clientId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "oauth2Record_clientSecret": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "oauth2Record_scope": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "oauth2Record_accessTokenUrl": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "oauth2Record_seleniumCreds": types.Boolean(),
        "tags_count": types.Integer(),
        "tags_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "comments_count": types.Integer(),
        "comments_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdDate": types.DateTime(),
        "updatedDate": types.DateTime(),
        "createdBy_id": types.Integer(),
        "createdBy_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdBy_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_id": types.Integer(),
        "updatedBy_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_firstName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedBy_lastName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # REDACT PASSWORDS IN serverRecord_fields_list
    # and formRecord_fields_list:
    REDACT_FIELDS = ["serverRecord_fields_list", "formRecord_fields_list"]
    for field in REDACT_FIELDS:
        for auth in authRecords:
            if getattr(auth, field):
                for record in getattr(auth, field):
                    record.redact_password()

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(record) for record in authRecords])

    # Drop any columns that we parsed out:
    df.drop(
        columns=[
            "owner",
            "formRecord",
            "serverRecord",
            "oauth2Record",
            "tags",
            "comments",
            "createdBy",
            "updatedBy",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)


def upload_was_findings(
    findings: BaseList,
    cnxn: Connection,
    table_name: str = "was_findings",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```was.get_findings```
    or ```was.get_findings_verbose``` to a SQL database.

    Args:
        findings (BaseList): A BaseList of WebAppFinding objects.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to.
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.Integer(),
        "uniqueId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "qid": types.Integer(),
        "detectionScore": types.Integer(),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "type": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "potential": types.Boolean(),
        "findingType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "group": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "severity": types.SmallInteger(),
        "url": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "firstDetectedDate": types.DateTime(),
        "lastDetectedDate": types.DateTime(),
        "lastTestedDate": types.DateTime(),
        "fixedDate": types.DateTime(),
        "timesDetected": types.Integer(),
        "webApp_id": types.Integer(),
        "webApp_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "webApp_url": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isIgnored": types.Boolean(),
        "ignoredReason": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ignoredBy": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ignoredDate": types.DateTime(),
        "ignoredComment": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "reactivatedDate": types.DateTime(),
        "reactivateIn": types.Integer(),
        "param": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "cwe_count": types.Integer(),
        "cwe_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "owasp_count": types.Integer(),
        "owasp_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "accessPath_count": types.Integer(),
        "accessPath_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "payloads_count": types.Integer(),
        "payloads_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "resultList_count": types.Integer(),
        "resultList_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "patch": types.Integer(),
        "cvssV3_base": types.Float(),
        "cvssV3_impact": types.Float(),
        "cvssV3_attackVector": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "severityComment": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "editedSeverityUser": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "history_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "wasc_count": types.Integer(),
        "wasc_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_protocol": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_virtualhost": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_ip": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_port": types.Integer(),
        "sslData_result": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_list": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_certificateFingerprint": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "sslData_flags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updatedDate": types.DateTime(),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(finding) for finding in findings])

    # Drop any columns that we parsed out:
    df.drop(
        columns=[
            "webApp",
            "cwe",
            "owasp",
            "resultList",
            "cvssV3",
            "history",
            "wasc",
            "sslData",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)



"""
tag.py - contains the CloudTag and Tag dataclasses for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *
from datetime import datetime
from uuid import UUID


@dataclass
class Tag:
    """
    Tag - represents a single tag in a TagList.

    This class is used to represent a single tag in a TagList, aka a tag on an asset.
    """

    TAG_ID: Union[str, int, UUID] = field(
        metadata={"description": "The ID of the tag."}
    )
    NAME: str = field(metadata={"description": "The name of the tag."})

    def __post_init__(self):
        # check that id is a string or int:
        if not isinstance(self.TAG_ID, (str, int, UUID)):
            raise TypeError(f"Tag ID must be a string or int, not {type(self.TAG_ID)}")

        # cast ID to an int if it is a string
        if isinstance(self.TAG_ID, str):
            self.TAG_ID = int(self.TAG_ID)

    def __str__(self) -> str:
        return self.NAME

    def __contains__(self, item):
        # see if it was found in the name or id:
        return item in self.TAG_ID or item in self.NAME

    def copy(self):
        return Tag(TAG_ID=self.TAG_ID, NAME=self.NAME)

    def is_id(self, id: int):
        return self.TAG_ID == id

    def is_name(self, name: str):
        return self.NAME == name

    def __iter__(self):
        yield self.TAG_ID
        yield self.NAME

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a Tag object from a dictionary.

        This function is used to create a Tag object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"TAG_ID", "NAME"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        # cast ID to an int if it is a string
        if isinstance(data["TAG_ID"], str):
            data["TAG_ID"] = int(data["TAG_ID"])

        return cls(**data)


@dataclass
class CloudTag:
    """
    CloudTag - represents a single tag in a CloudTagList.

    This class is used to represent a single tag in a CloudTagList, aka a tag on a cloud asset.
    """

    NAME: str = field(metadata={"description": "The name of the tag."})
    VALUE: str = field(metadata={"description": "The value of the tag."})
    LAST_SUCCESS_DATE: Optional[Union[str, datetime]] = field(
        default=None,
        metadata={"description": "The last successful date of the tag."},
    )

    def __post_init__(self):
        # check that the last success date is a string or datetime:
        if self.LAST_SUCCESS_DATE is not None and not isinstance(
            self.LAST_SUCCESS_DATE, (str, datetime)
        ):
            raise TypeError(
                f"Last success date must be a string or datetime, not {type(self.LAST_SUCCESS_DATE)}"
            )

        # cast last success date to a datetime if it is a string
        if isinstance(self.LAST_SUCCESS_DATE, str):
            self.LAST_SUCCESS_DATE = datetime.fromisoformat(self.LAST_SUCCESS_DATE)

    def __str__(self) -> str:
        return f"{self.NAME}:{self.VALUE}"

    def __dict__(self):
        return {
            "NAME": self.NAME,
            "VALUE": self.VALUE,
            "LAST_SUCCESS_DATE": self.LAST_SUCCESS_DATE,
        }

    def __contains__(self, item):
        # see if it was found in the name or value:
        return item in self.NAME or item in self.VALUE

    def copy(self):
        return CloudTag(NAME=self.NAME, VALUE=self.VALUE)

    def is_name(self, name: str):
        return self.NAME == name

    def is_value(self, value: str):
        return self.VALUE == value

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a CloudTag object from a dictionary.

        This function is used to create a CloudTag object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"NAME", "VALUE"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)



"""
Contains the get_remediation_activities function for Totalcloud
"""

from typing import Union, Literal

from ..base.call_api import call_api
from ..base.base_list import BaseList
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.RemediationActivity import RemediationActivity


def get_remediation_activities(
    auth: BasicAuth,
    provider: Literal["aws", "azure"],
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> BaseList[dict]:
    """
    Pull remediation activity by cloud provider

    Args:
        auth (BasicAuth): The authentication object.
        provider (Literal['aws', 'azure']): The cloud provider of the resource.
        page_count (Union[int, 'all']): The number of pages to retrieve. If 'all', all pages are retrieved.

    ## Kwargs:
        - filter (str): Filter resources by providing a QQL [query](https://docs.qualys.com/en/cloudview/latest/search_tips/search_remediation_activity.htm?rhhlterm=search%20remediation%20activity%20activities&rhsearch=Search%20for%20Remediation%20Activity)
        - pageNo (int): The ordered page to start retrieving resources from, or if page_count is 1, the page to retrieve.
        - pageSize (int): The number of resources to get per page.

    Returns:
        BaseList[dict]: A list of dictionaries containing the remediation activities. WARNING: I have not seen the response from this API, so I am not sure what the response will look like.
    """

    if (page_count != "all" and (not isinstance(page_count, int))) or (
        isinstance(page_count, int) and page_count < 1
    ):
        raise ValueError("page_count must be an integer >=1 or 'all'.")

    if provider.lower() not in ["aws", "azure"]:
        raise ValueError("Invalid provider. Must be 'aws' or 'azure'.")

    params = {
        "cloudType": provider.upper(),
    }

    # Add in kwargs to the params
    if kwargs:
        params.update(kwargs)

    pulled_pages = 0
    bl = BaseList()

    while True:
        # Call the API
        response = call_api(
            auth=auth,
            module="cloudview",
            endpoint="get_remediation_activities",
            params=params,
        )

        # Check the response
        if response.status_code != 200:
            raise QualysAPIError(f"{response.status_code}: {response.text}")

        j = response.json()

        if "errorCode" in j:
            raise QualysAPIError(f"{j['errorCode']}, {j['message']}")

        if "content" not in j.keys() or not j["pageable"].get("empty"):
            print("No content in response")
            break

        data = j["content"]
        if isinstance(data, dict):
            data = [data]
        for item in data:
            bl.append(RemediationActivity.from_dict(item))

        # Check if we need to pull more pages
        if page_count != "all":
            pulled_pages += 1
            if pulled_pages >= page_count:
                print("Reached page limit of", page_count)
                break

    return bl



"""
Contains user-facing functions for interacting with authentication records in WAS
"""

from typing import Union, Literal, List, Dict, Optional

from .data_classes.WebAppAuthRecord import WebAppAuthRecord
from .base.auth_record_service_requests import (
    create_service_request,
    unparse_to_xml_str,
)
from .base.web_app_service_requests import build_service_request
from .base.parse_kwargs import validate_kwargs
from .base.web_app_service_requests import validate_response
from ..base.call_api import call_api
from ..auth.basic import BasicAuth
from ..exceptions.Exceptions import QualysAPIError
from ..base.base_list import BaseList


def call_auth_api(
    auth: BasicAuth, endpoint: str, payload: dict
) -> Union[int, WebAppAuthRecord]:
    """
    Call a Qualys WAS API auth record endpoint and return the parsed response. This is
    a backend function and should not be called directly.

    Args:
        auth (BasicAuth): The authentication object.
        endpoint (str): The API endpoint to call.
        payload (dict): The payload to send to the API.

    Returns:
        Union[int, WebAppAuthRecord]: The parsed response from the API.
    """

    match endpoint:
        case "count_authentication_records":
            params = {"placeholder": "count", "webappAuthRecordId": ""}
        case "get_authentication_records":
            params = {"placeholder": "search", "webappAuthRecordId": ""}
        case "get_authentication_record_details":
            params = {
                "placeholder": "get",
                "webappAuthRecordId": payload.pop("recordId"),
            }
        case "create_authentication_record":
            params = {"placeholder": "create", "webappAuthRecordId": ""}
        case "delete_authentication_record":
            params = {"placeholder": "delete", "webappAuthRecordId": ""}
        case _:
            raise ValueError(f"Invalid endpoint: {endpoint}")

    response = call_api(
        auth=auth,
        override_method=(
            "GET" if endpoint == "get_authentication_record_details" else "POST"
        ),
        module="was",
        endpoint="call_auth_api",
        params=params,
        payload=payload,
        headers={"Content-Type": "text/xml"},
    )

    return validate_response(response)


def count_authentication_records(auth: BasicAuth, **kwargs) -> int:
    """
    Return a count of auth records in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.

    ## Kwargs:

        - id (Union[str, int]): Authentication object ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Auth object name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - tags (Union[str, int]): Tag ID.
        - tags_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tags filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the Auth object was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the Auth object was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isUsed (bool): Whether the Auth object is in use.
        - isUsed (Literal["EQUALS", "NOT EQUALS"]): Operator for the isUsed filter.
        - lastScan_authStatus (Literal["NONE", "NOT_USED", "PARTIAL", "FAILED", "SUCCESSFUL"]): Status of the last scan.
        - lastScan_authStatus_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - contents (Literal["FORM_STANDARD", "FORM_CUSTOM", "FORM_SELENIUM", "SERVER_BASIC", "SERVER_DIGEST", "SERVER_NTLM", "CERTIFICATE", "OAUTH2_AUTH_CODE", "OAUTH2_IMPLICIT", "OAUTH2_PASSWORD", "OAUTH2_CLIENT_CREDS"]): Auth object contents.
        - contents_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the contents filter.

    Returns:
        int: The count of Auth objects that match the filters.
    """

    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="count_authentication_records", **kwargs)
        payload = build_service_request(**kwargs)

    # Make the API call:
    parsed = call_auth_api(auth, "count_authentication_records", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    return int(serviceResponse.get("count"))


def get_authentication_records(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> object:
    """
    Get a list of auth records in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        - id (Union[str, int]): Authentication object ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Auth object name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - tags (Union[str, int]): Tag ID.
        - tags_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tags filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the Auth object was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the Auth object was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isUsed (bool): Whether the Auth object is in use.
        - isUsed (Literal["EQUALS", "NOT EQUALS"]): Operator for the isUsed filter.
        - lastScan_authStatus (Literal["NONE", "NOT_USED", "PARTIAL", "FAILED", "SUCCESSFUL"]): Status of the last scan.
        - lastScan_authStatus_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - contents (Literal["FORM_STANDARD", "FORM_CUSTOM", "FORM_SELENIUM", "SERVER_BASIC", "SERVER_DIGEST", "SERVER_NTLM", "CERTIFICATE", "OAUTH2_AUTH_CODE", "OAUTH2_IMPLICIT", "OAUTH2_PASSWORD", "OAUTH2_CLIENT_CREDS"]): Auth object contents.
        - contents_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the contents filter.

    Returns:
        object: The auth records that match the filters.
    """

    # Ensure that page_count is either the string 'all' or an integer >= 1:
    if isinstance(page_count, int) and page_count < 1:
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")
    elif isinstance(page_count, str) and page_count != "all":
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")

    pageNo = 0
    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="get_authentication_records", **kwargs)
        payload = build_service_request(**kwargs)

    appList = BaseList()

    while True:
        # Make the API call:
        parsed = call_auth_api(auth, "get_authentication_records", payload)

        # Parse the XML response:
        serviceResponse = parsed.get("ServiceResponse")
        if not serviceResponse:
            raise QualysAPIError("No ServiceResponse tag returned in the API response")

        if serviceResponse.get("responseCode") != "SUCCESS":
            raise QualysAPIError(
                f"API response returned error: {serviceResponse.get('responseCode')}"
            )

        if serviceResponse.get("count") == "0":
            print(f"No web applications found on page {pageNo}. Exiting.")
            break

        data = serviceResponse.get("data")

        if data.get("WebAppAuthRecord"):
            data = data.get("WebAppAuthRecord")

        if isinstance(data, dict):
            data = [data]

        for record in data:
            # Create the objects:
            appList.append(WebAppAuthRecord.from_dict(record))

        print(
            f"Retrieved {serviceResponse.get('count')} auth records on page {pageNo}. Running total: {len(appList)}"
        )

        pageNo += 1

        if page_count != "all" and pageNo >= page_count:
            print(f"Reached page_count limit. Returning {pageNo} page(s).")
            break

        # Check for pagination:
        if serviceResponse.get("hasMoreRecords") == "true":
            # Update the XML payload with the new Criteria:
            # <Criteria field="id" operator="GREATER">XXX</Criteria>
            kwargs["id.operator"] = "GREATER"
            kwargs["id"] = serviceResponse.get("lastId")
            payload = build_service_request(**kwargs)
        else:
            break

    return appList


def get_authentication_record_details(
    auth: BasicAuth, recordId: Union[int, str]
) -> WebAppAuthRecord:
    """
    Pull a single authentication record with all attributes populated.

    Args:
        auth (BasicAuth): The authentication object.
        recordId (Union[int, str]): The ID of the record to pull.

    Returns:
        WebAppAuthRecord: The authentication record.
    """
    # Make the API call:
    parsed = call_auth_api(
        auth, "get_authentication_record_details", {"recordId": recordId}
    )

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")
    if data.get("WebAppAuthRecord"):
        data = data.get("WebAppAuthRecord")
        return WebAppAuthRecord.from_dict(data)
    else:
        print(f"No data found for web application ID {recordId}. Exiting.")


def get_authentication_records_verbose(
    auth: BasicAuth, thread_count: int = 5, **kwargs
) -> BaseList[WebAppAuthRecord]:
    """
    Uses ```was.get_authentication_records()``` and ```was.get_authentication_record_details()``` to return a ```BaseList``` of ```WebAppAuthRecord```s with
    all attributes populated.

    This function is multi-threaded, placing all ```WebAppAuthRecords``` found
    from ```was.get_authentication_records()``` into a queue and then spawning threads to pull the details one
    by one.

    The details threads wait for work to be added to the queue and then pull
    the details for each ```WebAppAuthRecord.id```.

    Args:
        auth (BasicAuth): The authentication object.
        thread_count (int): The number of threads to spawn. Defaults to 5.

    ## Kwargs:

        - id (Union[str, int]): Authentication object ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Auth object name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - tags (Union[str, int]): Tag ID.
        - tags_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tags filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the Auth object was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the Auth object was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isUsed (bool): Whether the Auth object is in use.
        - isUsed (Literal["EQUALS", "NOT EQUALS"]): Operator for the isUsed filter.
        - lastScan_authStatus (Literal["NONE", "NOT_USED", "PARTIAL", "FAILED", "SUCCESSFUL"]): Status of the last scan.
        - lastScan_authStatus_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - contents (Literal["FORM_STANDARD", "FORM_CUSTOM", "FORM_SELENIUM", "SERVER_BASIC", "SERVER_DIGEST", "SERVER_NTLM", "CERTIFICATE", "OAUTH2_AUTH_CODE", "OAUTH2_IMPLICIT", "OAUTH2_PASSWORD", "OAUTH2_CLIENT_CREDS"]): Auth object contents.
        - contents_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the contents filter.

    Returns:
        BaseList[WebAppAuthRecord]: The list of auth records.
    """

    # Check thread_count for a valid value:
    if not isinstance(thread_count, int) or thread_count < 1:
        raise ValueError("thread_count must be an integer >= 1.")

    # Set up Queue and List, with some cheeky as-needed imports:
    from queue import Queue
    from threading import Thread, Lock, current_thread

    q = Queue()
    authList = BaseList()
    LOCK = Lock()
    threads = []

    # Get the auth records:
    print(f"({current_thread().name}) Getting base auth record list...")
    authrecords = get_authentication_records(auth, **kwargs)

    print(
        f"({current_thread().name}) Pulled {len(authrecords)} auth records. Starting {thread_count} thread(s) for details pull.."
    )

    # Add the auth records to the queue:
    for record in authrecords:
        q.put(record)

    def worker():
        while True:
            try:
                # Exit condition 1: Queue is empty
                if q.empty():
                    with LOCK:
                        print(
                            f"({current_thread().name}) Queue is empty. Thread exiting."
                        )
                        break

                authrecord = q.get()
                # Exit condition 2: authrecord is None (because Queue is empty)
                if not authrecord:
                    with LOCK:
                        print(
                            f"({current_thread().name}) Queue is empty. Thread exiting."
                        )
                        q.task_done()
                    break

                details = get_authentication_record_details(auth, authrecord.id)
                authList.append(details)
                q.task_done()
                with LOCK:
                    if len(authList) % 10 == 0:
                        print(
                            f"({current_thread().name}) Pulled {len(authList)} auth record details so far..."
                        )

            except Exception as e:
                with LOCK:
                    print(
                        f"[ERROR - THREAD EXITING] ({current_thread().name}) Error: {e}"
                    )
                q.task_done()
                break

    # Start the threads:
    for i in range(thread_count):
        t = Thread(target=worker)
        threads.append(t)
        t.start()

    # Wait for the threads to finish:
    for t in threads:
        t.join()

    print(f"Pulled {len(authList)} auth record details.")
    return authList


"""
format_tag_list takes in a list of ints 
and returns a properly structured list 
of dicts for the API payload.
"""
format_tag_list = lambda tags: [{"id": tag} for tag in tags]


def create_authentication_record(
    auth: BasicAuth,
    name: str,
    recordType: Literal["formRecord", "serverRecord", "oauth2Record"],
    fields: List[Dict[str, Union[str, bool]]] = None,
    subType: Optional[
        Literal[
            "STANDARD",
            "CUSTOM",
            "SELENIUM",
            "BASIC",
            "DIGEST",
            "NTLM",
            "AUTH_CODE",
            "IMPLICIT",
            "PASSWORD",
            "CLIENT_CREDS",
        ]
    ] = None,
    **kwargs,
) -> WebAppAuthRecord:
    """
    Create a new authentication record in WAS

    Args:
        auth (BasicAuth): The authentication object.
        name (str): The name of the authentication record.
        recordType (Literal["formRecord", "serverRecord", "oauth2Record"]): The type of authentication record to create.
        subType (Optional[Literal["STANDARD", "CUSTOM", "SELENIUM", "BASIC", "DIGEST", "NTLM", "AUTH_CODE", "IMPLICIT", "PASSWORD", "CLIENT_CREDS"]]): The subtype of the authentication record to create.
        fields (List[Dict[str, Union[str, bool]]]): The fields to include in the authentication record. Each field should be a dictionary with the keys 'name', 'value', and optionally 'secured'.

    ## Kwargs:

        - tags (List[int]): The tag ID numbers to associate with the authentication record as a list. Default is an empty list.
        - comments (Union[List[str], str]): Comments to associate with the authentication record. Default is an empty list.
        - sslOnly (bool): Whether the authentication record should only be used over encrypted connections. Default is False.
        - authVault (bool): Whether the authentication record should be stored in the Qualys Vault. Default is False.
        - certificate (Dict[str, str]): A dictionary containing the certificate information for the authentication record. Valid for serverRecord records. Example: certificate={"name": "certname", "contents": "certcontents", "passphrase": "certpassphrase"}}.
        - grantType (Literal["AUTH_CODE", "IMPLICIT", "PASSWORD", "CLIENT_CREDS"]): The grant type for the OAuth2 record. Default is None.
        - accessTokenUrl (str): The access token URL for OAuth2 records.
        - redirectUrl (str): The redirect URL for OAuth2 records.
        - username (str): The username for OAuth2/Server records.
        - password (str): The password for OAuth2/Server records.
        - clientId (str): The client ID for OAuth2 records.
        - clientSecret (str): The client secret for OAuth2 records.
        - scope (str): The scope for OAuth2 records.
        - accessTokenExpiredMsgPattern (str): The access token expired message pattern for OAuth2 records.
        - seleniumScript (Dict[str, str]): The Selenium script details for formRecord and OAuth2 records. Example: {"name": "ImplicitScript", "data": "implicit script data", "regex": "implicit regex pattern"}.
        - seleniumCreds (bool): Whether Selenium credentials are used. Default is None.

    Returns:
        WebAppAuthRecord: The created authentication record.
    """

    # Check validity of recordType:
    if recordType not in ["formRecord", "serverRecord", "oauth2Record"]:
        raise ValueError(
            "recordType must be one of 'formRecord', 'serverRecord', or 'oauth2Record'."
        )

    # Check validity of subType:
    acceptable_subtypes_by_recordType = {
        "formRecord": ["STANDARD", "CUSTOM", "SELENIUM"],
        "serverRecord": ["BASIC", "DIGEST", "NTLM", None],
        "oauth2Record": ["AUTH_CODE", "IMPLICIT", "PASSWORD", "CLIENT_CREDS"],
    }
    if subType and subType.upper() not in acceptable_subtypes_by_recordType[recordType]:
        raise ValueError(
            f"subType must be one of {acceptable_subtypes_by_recordType[recordType]} for a {recordType}, not {subType}."
        )

    # Ensure tags is a list
    if not isinstance(kwargs.get("tags", []), list):
        kwargs["tags"] = [kwargs["tags"]]

    # Ensure comments is a list
    if isinstance(kwargs.get("comments", []), str):
        kwargs["comments"] = [kwargs["comments"]]

    # Build the payload:
    payload = {
        "name": name,
        recordType: {
            "type" if recordType != "oauth2Record" else "grantType": subType,
            "sslOnly": kwargs.get("sslOnly"),
            "authVault": kwargs.get("authVault"),
            "fields": {"set": fields},
        },
        "tags": {"set": format_tag_list(kwargs.get("tags", []))},
        "comments": {"set": kwargs.get("comments", [])},
    }

    # Add optional fields based on recordType
    if recordType == "serverRecord" and "certificate" in kwargs:
        payload[recordType]["certificate"] = kwargs["certificate"]

    if recordType == "oauth2Record":
        oauth2_fields = [
            "grantType",
            "accessTokenUrl",
            "redirectUrl",
            "username",
            "password",
            "clientId",
            "clientSecret",
            "scope",
            "accessTokenExpiredMsgPattern",
            "seleniumScript",
            "seleniumCreds",
        ]
        for field in oauth2_fields:
            if field in kwargs:
                payload[recordType][field] = kwargs[field]

    # Create the service request
    xml_payload = unparse_to_xml_str(
        create_service_request(data={"WebAppAuthRecord": payload})
    )

    # Make the API call:
    parsed = call_auth_api(
        auth, "create_authentication_record", {"_xml_data": xml_payload}
    )

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")
    if data.get("WebAppAuthRecord"):
        data = data.get("WebAppAuthRecord")
        return WebAppAuthRecord.from_dict(data)
    else:
        print("No data found. Exiting.")


def delete_authentication_record(auth: BasicAuth, **kwargs) -> list[str]:
    """
    Delete authentication records out of Qualys WAS.

    To delete a single record, use kwarg 'id' set to
    a single int/str.

    To delete multiple assets,
    set 'id' to a list of int/str, or use the other
    kwargs/operators listed below.

    If both 'id' and other kwargs are provided, the 'id'
    kwarg will take precedence.

    Args:
        auth (BasicAuth): The authentication object.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - tags (Union[str, int]): Tag ID.
        - tags_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tags filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - lastScan_authStatus (Literal["NONE", "NOT_USED", "SUCCESSFUL", "FAILED", "PARTIAL"]): Status of the last scan.
        - lastScan_authStatus_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - isUsed (bool): Whether the web application is in use.
        - isUsed_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isUsed filter.
        - contents (Literal["FORM_STANDARD", "FORM_CUSTOM", "FORM_SELENIUM", "SERVER_BASIC", "SERVER_DIGEST", "SERVER_NTLM", "CERTIFICATE", "OAUTH2_AUTH_CODE", "OAUTH2_IMPLICIT", "OAUTH2_PASSWORD", "OAUTH2_CLIENT_CREDS"]): Web application contents.
        - contents_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the contents filter.

    Returns:
        list[str]: A list of IDs that were successfully deleted.
    """

    # Make sure user passed in a kwarg:
    if not kwargs:
        raise ValueError("No kwargs provided. Please provide at least one kwarg.")

    TO_STR = [
        "id",
        "tags_id",
    ]

    for key in TO_STR:
        if key in kwargs:
            kwargs[key] = str(kwargs[key])

    # Validate the kwargs:
    kwargs = validate_kwargs(endpoint="delete_authentication_record", **kwargs)

    # Build the XML payload:
    payload = build_service_request(**kwargs)

    # Make the API call:
    parsed = call_auth_api(auth, "delete_authentication_record", payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    if serviceResponse.get("count") == "0":
        print(f"No auth records found. Exiting.")
        return []

    deleted = []
    data = serviceResponse.get("data")
    if data.get("WebAppAuthRecord"):
        data = data.get("WebAppAuthRecord")
        if isinstance(data, dict):
            data = [data]
        for record in data:
            deleted.append(record)

    return deleted



"""
Work with Qualys WAS findings
"""

from typing import Union, List, Dict, Literal

from .data_classes.Finding import WASFinding
from .base.parse_kwargs import validate_kwargs
from .base.web_app_service_requests import build_service_request
from ..auth.basic import BasicAuth
from ..base.call_api import call_api
from .base.web_app_service_requests import validate_response
from ..exceptions.Exceptions import QualysAPIError
from ..base.base_list import BaseList


def call_findings_api(
    auth: BasicAuth, endpoint: str, payload: dict
) -> Union[int, BaseList[WASFinding], WASFinding]:
    """
    Call the Qualys WAS findings API

    Args:
        auth (BasicAuth): The authentication object
        endpoint (str): The endpoint to call
        payload (dict): The payload to send

    Returns:
        Union[int, BaseList[WASFinding], WASFinding]: The response from the API
    """

    match endpoint:
        case "count_findings":
            params = {"placeholder": "count", "findingId": ""}
        case "get_findings":
            params = {"placeholder": "search", "findingId": ""}
        case "get_finding_details":
            params = {"placeholder": "get", "findingId": payload.pop("findingId")}
        case _:
            raise ValueError("Invalid endpoint: {endpoint}")

    response = call_api(
        auth=auth,
        override_method="GET" if endpoint == "get_finding_details" else "POST",
        module="was",
        endpoint="call_findings_api",
        payload=payload,
        params=params,
        headers={"Content-Type": "text/xml"},
    )

    return validate_response(response)


def count_findings(auth: BasicAuth, **kwargs) -> int:
    """
    Count how many findings are in the Qualys WAS module
    according to the filters provided

    Args:
        auth (BasicAuth): The authentication object

    ## Kwargs:

        - id (int): The finding ID
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - uniqueId (str): The unique ID of the finding
        - qid (int): The Qualys ID of the finding
        - qid_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the QID filter.
        - name (str): The name of the finding
        - name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the name filter.
        - type (Literal["VULNERABILITY", "SENSITIVE_CONTENT", "INFORMATION_GATHERED"]): The type of the finding
        - type_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the type filter.
        - url (str): The URL of the finding
        - url_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the URL filter.
        - webApp_tags_id (int): The ID of the web application tag on a web application
        - webApp_tags_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the webApp_tags_id filter.
        - webApp_tags_name (str): The name of a web application tag on a web application
        - webApp_tags_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the webApp_tags_name filter.
        - status (Literal["NEW", "ACTIVE", "REOPENED", "PROTECTED", "FIXED"]): The status of the finding
        - status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the status filter.
        - patch (int): Use WAF to protect against vulnerabilities by installing virtual patches
        - patch_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the patch filter.
        - webApp_id (int): The ID of the web application
        - webApp_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the webApp_id filter.
        - webApp_name (str): The name of the web application
        - webApp_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the webApp_name filter.
        - severity (Literal[1, 2, 3, 4, 5]): The severity of the finding
        - severity_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the severity filter.
        - externalRef (str): The external reference of the finding
        - externalRef_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the externalRef filter.
        - ignoredDate (str): The date the finding was ignored as a UTC timestamp
        - ignoredDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ignoredDate filter.
        - ignoredReason (Literal["FALSE_POSITIVE", "RISK_ACCEPTED", "NOT_APPLICABLE"]): The reason the finding was ignored
        - ignoredReason_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the ignoredReason filter.
        - group (Literal["XSS", "SQL", "INFO", "PATH", "CC", "SSN_US", "CUSTOM"]): The group of the finding
        - group_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the group filter.
        - owasp_name (str): The OWASP name of the finding
        - owasp_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the owasp_name filter.
        - owasp_code (int): The OWASP code of the finding
        - owasp_code_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the owasp_code filter.
        - wasc_name (str): The WASC name of the finding
        - wasc_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the wasc_name filter.
        - wasc_code (int): The WASC code of the finding
        - wasc_code_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the wasc_code filter.
        - cwe_id (int): The CWE ID of the finding
        - cwe_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the cwe_id filter.
        - firstDetectedDate (str): The date the finding was first detected as a UTC timestamp
        - firstDetectedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the firstDetectedDate filter.
        - lastDetectedDate (str): The date the finding was last detected as a UTC timestamp
        - lastDetectedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the lastDetectedDate filter.
        - lastTestedDate (str): The date the finding was last tested as a UTC timestamp
        - lastTestedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the lastTestedDate filter.
        - timesDetected (int): The number of times the finding was detected
        - timesDetected_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the timesDetected filter.

    Returns:
        int: Number of findings
    """

    payload = None

    if kwargs:
        for key, value in kwargs.items():
            # cast all to string:
            kwargs[key] = str(value)
        kwargs = validate_kwargs(endpoint="count_findings", **kwargs)

        payload = build_service_request(**kwargs)

    # Make the API call
    parsed = call_findings_api(auth=auth, endpoint="count_findings", payload=payload)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response code was not SUCCESS: {serviceResponse.get('responseCode')}"
        )

    return int(serviceResponse.get("count"))


def get_findings(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> BaseList[WASFinding]:
    """
    Get a list of findings from Qualys WAS according
    to the filters provided

    Args:
        auth (BasicAuth): The authentication object
        page_count (Union[int, "all"]): The number of pages to retrieve. Default is "all"

    ## Kwargs:

        - id (int): The finding ID
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - uniqueId (str): The unique ID of the finding
        - qid (int): The Qualys ID of the finding
        - qid_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the QID filter.
        - name (str): The name of the finding
        - name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the name filter.
        - type (Literal["VULNERABILITY", "SENSITIVE_CONTENT", "INFORMATION_GATHERED"]): The type of the finding
        - type_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the type filter.
        - url (str): The URL of the finding
        - url_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the URL filter.
        - webApp_tags_id (int): The ID of the web application tag on a web application
        - webApp_tags_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the webApp_tags_id filter.
        - webApp_tags_name (str): The name of a web application tag on a web application
        - webApp_tags_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the webApp_tags_name filter.
        - status (Literal["NEW", "ACTIVE", "REOPENED", "PROTECTED", "FIXED"]): The status of the finding
        - status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the status filter.
        - patch (int): Use WAF to protect against vulnerabilities by installing virtual patches
        - patch_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the patch filter.
        - webApp_id (int): The ID of the web application
        - webApp_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the webApp_id filter.
        - webApp_name (str): The name of the web application
        - webApp_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the webApp_name filter.
        - severity (Literal[1, 2, 3, 4, 5]): The severity of the finding
        - severity_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the severity filter.
        - externalRef (str): The external reference of the finding
        - externalRef_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the externalRef filter.
        - ignoredDate (str): The date the finding was ignored as a UTC timestamp
        - ignoredDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ignoredDate filter.
        - ignoredReason (Literal["FALSE_POSITIVE", "RISK_ACCEPTED", "NOT_APPLICABLE"]): The reason the finding was ignored
        - ignoredReason_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the ignoredReason filter.
        - group (Literal["XSS", "SQL", "INFO", "PATH", "CC", "SSN_US", "CUSTOM"]): The group of the finding
        - group_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the group filter.
        - owasp_name (str): The OWASP name of the finding
        - owasp_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the owasp_name filter.
        - owasp_code (int): The OWASP code of the finding
        - owasp_code_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the owasp_code filter.
        - wasc_name (str): The WASC name of the finding
        - wasc_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the wasc_name filter.
        - wasc_code (int): The WASC code of the finding
        - wasc_code_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the wasc_code filter.
        - cwe_id (int): The CWE ID of the finding
        - cwe_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the cwe_id filter.
        - firstDetectedDate (str): The date the finding was first detected as a UTC timestamp
        - firstDetectedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the firstDetectedDate filter.
        - lastDetectedDate (str): The date the finding was last detected as a UTC timestamp
        - lastDetectedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the lastDetectedDate filter.
        - lastTestedDate (str): The date the finding was last tested as a UTC timestamp
        - lastTestedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the lastTestedDate filter.
        - timesDetected (int): The number of times the finding was detected
        - timesDetected_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the timesDetected filter.
        - fixedDate (str): The date the finding was fixed as a UTC timestamp
        - fixedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the fixedDate filter.
        - verbose (bool): Whether to return verbose output

    Returns:
        BaseList[WASFinding]: A list of WASFinding objects
    """

    if page_count != "all" and not (isinstance(page_count, int) or page_count < 0):
        raise ValueError("page_count must be 'all' or a positive integer")

    pageNo = 0
    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="get_findings", **kwargs)
        payload = build_service_request(**kwargs)

    findingList = BaseList()

    while True:
        # Make the API call:
        parsed = call_findings_api(auth, "get_findings", payload)

        # Parse the XML response:
        serviceResponse = parsed.get("ServiceResponse")
        if not serviceResponse:
            raise QualysAPIError("No ServiceResponse tag returned in the API response")

        if serviceResponse.get("responseCode") != "SUCCESS":
            raise QualysAPIError(
                f"API response returned error: {serviceResponse.get('responseCode')}"
            )

        if serviceResponse.get("count") == "0":
            print(f"No findings found on page {pageNo}. Exiting.")
            break

        data = serviceResponse.get("data")

        if data.get("Finding"):
            data = data.get("Finding")

        if isinstance(data, dict):
            data = [data]

        for finding in data:
            # Create the objects:
            findingList.append(WASFinding.from_dict(finding))

        print(
            f"Retrieved {serviceResponse.get('count')} WAS findings on page {pageNo}. Running total: {len(findingList)}"
        )

        pageNo += 1

        if page_count != "all" and pageNo >= page_count:
            print(f"Reached page_count limit. Returning {pageNo} page(s).")
            break

        # Check for pagination:
        if serviceResponse.get("hasMoreRecords") == "true":
            # Update the XML payload with the new Criteria:
            # <Criteria field="id" operator="GREATER">XXX</Criteria>
            kwargs["id.operator"] = "GREATER"
            kwargs["id"] = serviceResponse.get("lastId")
            payload = build_service_request(**kwargs)
        else:
            break

    return findingList


def get_finding_details(auth: BasicAuth, findingId: Union[str, int]) -> WASFinding:
    """
    Pull all details of a single finding from Qualys WAS
    by its finding # or unique ID

    Args:
        auth (BasicAuth): The authentication object
        findingId (Union[str, int]): The finding number or unique ID of the finding

    Returns:
        WASFinding: The WASFinding object
    """

    if not isinstance(findingId, (str, int)):
        raise ValueError("findingId must be a string or integer")

    # Make the API call
    parsed = call_findings_api(auth, "get_finding_details", {"findingId": findingId})

    # Parse the XML response:
    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    data = serviceResponse.get("data")

    if data.get("Finding"):
        data = data.get("Finding")

    return WASFinding.from_dict(data)


def get_findings_verbose(
    auth: BasicAuth, thread_count: int = 5, **kwargs
) -> BaseList[WASFinding]:
    """
    Uses ```was.get_findings()``` and ```was.get_finding_details()``` to return a ```BaseList``` of ```WASFinding```s with
    all attributes populated.

    This function is multi-threaded, placing all ```WASFinding```s found
    from ```was.get_findings()``` into a queue and then spawning threads to pull the details one
    by one.

    The details threads wait for work to be added to the queue and then pull
    the details for each ```WASFinding.id```.

    Args:
        auth (BasicAuth): The authentication object.
        thread_count (int): The number of threads to spawn. Defaults to 5.

    ## Kwargs:

        - id (int): The finding ID
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - uniqueId (str): The unique ID of the finding
        - qid (int): The Qualys ID of the finding
        - qid_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the QID filter.
        - name (str): The name of the finding
        - name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the name filter.
        - type (Literal["VULNERABILITY", "SENSITIVE_CONTENT", "INFORMATION_GATHERED"]): The type of the finding
        - type_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the type filter.
        - url (str): The URL of the finding
        - url_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the URL filter.
        - webApp_tags_id (int): The ID of the web application tag on a web application
        - webApp_tags_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the webApp_tags_id filter.
        - webApp_tags_name (str): The name of a web application tag on a web application
        - webApp_tags_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the webApp_tags_name filter.
        - status (Literal["NEW", "ACTIVE", "REOPENED", "PROTECTED", "FIXED"]): The status of the finding
        - status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the status filter.
        - patch (int): Use WAF to protect against vulnerabilities by installing virtual patches
        - patch_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the patch filter.
        - webApp_id (int): The ID of the web application
        - webApp_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the webApp_id filter.
        - webApp_name (str): The name of the web application
        - webApp_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the webApp_name filter.
        - severity (Literal[1, 2, 3, 4, 5]): The severity of the finding
        - severity_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the severity filter.
        - externalRef (str): The external reference of the finding
        - externalRef_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the externalRef filter.
        - ignoredDate (str): The date the finding was ignored as a UTC timestamp
        - ignoredDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ignoredDate filter.
        - ignoredReason (Literal["FALSE_POSITIVE", "RISK_ACCEPTED", "NOT_APPLICABLE"]): The reason the finding was ignored
        - ignoredReason_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the ignoredReason filter.
        - group (Literal["XSS", "SQL", "INFO", "PATH", "CC", "SSN_US", "CUSTOM"]): The group of the finding
        - group_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the group filter.
        - owasp_name (str): The OWASP name of the finding
        - owasp_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the owasp_name filter.
        - owasp_code (int): The OWASP code of the finding
        - owasp_code_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the owasp_code filter.
        - wasc_name (str): The WASC name of the finding
        - wasc_name_operator (Literal["EQUALS", "NOT EQUALS", "CONTAINS"]): Operator for the wasc_name filter.
        - wasc_code (int): The WASC code of the finding
        - wasc_code_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the wasc_code filter.
        - cwe_id (int): The CWE ID of the finding
        - cwe_id_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the cwe_id filter.
        - firstDetectedDate (str): The date the finding was first detected as a UTC timestamp
        - firstDetectedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the firstDetectedDate filter.
        - lastDetectedDate (str): The date the finding was last detected as a UTC timestamp
        - lastDetectedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the lastDetectedDate filter.
        - lastTestedDate (str): The date the finding was last tested as a UTC timestamp
        - lastTestedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the lastTestedDate filter.
        - timesDetected (int): The number of times the finding was detected
        - timesDetected_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the timesDetected filter.
        - fixedDate (str): The date the finding was fixed as a UTC timestamp
        - fixedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the fixedDate filter.

    Returns:
        BaseList[WASFinding]: The list of WASFinding objects.
    """

    # Check thread_count for a valid value:
    if not isinstance(thread_count, int) or thread_count < 1:
        raise ValueError("thread_count must be an integer >= 1.")

    # Set up Queue and List, with some cheeky as-needed imports:
    from queue import Queue
    from threading import Thread, Lock, current_thread

    q = Queue()
    findingList = BaseList()
    LOCK = Lock()
    threads = []

    # Get the findings:
    print(f"({current_thread().name}) Getting base finding list...")
    findings = get_findings(auth, **kwargs)

    print(
        f"({current_thread().name}) Pulled {len(findings)} findings. Starting {thread_count} thread(s) for details pull.."
    )

    # Add the findings to the queue:
    for record in findings:
        q.put(record)

    def worker():
        while True:
            try:
                # Exit condition 1: Queue is empty
                if q.empty():
                    with LOCK:
                        print(
                            f"({current_thread().name}) Queue is empty. Thread exiting."
                        )
                        break

                finding = q.get()
                # Exit condition 2: authrecord is None (because Queue is empty)
                if not finding:
                    with LOCK:
                        print(
                            f"({current_thread().name}) Queue is empty. Thread exiting."
                        )
                        q.task_done()
                    break

                details = get_finding_details(auth, finding.id)
                findingList.append(details)
                q.task_done()
                with LOCK:
                    if len(findingList) % 10 == 0:
                        print(
                            f"({current_thread().name}) Pulled {len(findingList)} finding details so far..."
                        )

            except Exception as e:
                with LOCK:
                    print(
                        f"[ERROR - THREAD EXITING] ({current_thread().name}) Error: {e}"
                    )
                q.task_done()
                break

    # Start the threads:
    for i in range(thread_count):
        t = Thread(target=worker)
        threads.append(t)
        t.start()

    # Wait for the threads to finish:
    for t in threads:
        t.join()

    print(f"Pulled {len(findingList)} finding details.")
    return findingList



"""
qualysdk - Qualys API SDK for Python

This package aims to make it easier to interact with the Qualys API across all of the different modules that Qualys provides.
"""

from .help import schema_query
from .auth import BasicAuth, TokenAuth
from .base.base_list import BaseList
from .base.csv_export import write_csv, write_excel

from .gav.uber import GAVUber
from . import gav

from .vmdr import query_kb, get_host_list, get_hld
from . import vmdr

from . import cloud_agent
from . import totalcloud
from . import cs
from . import was

from .sql import db_connect

# surprise!
__surprise__ = b"\xe2\x9c\xa8\xe2\x9c\xa8\xe2\x9c\xa8 Have a great day!".decode("utf-8")



"""
Qualysdk-updater is a CLI tool installed with 
the SDK to check for and install updates from 
PyPI.
"""

from argparse import ArgumentParser
from subprocess import run, check_call, DEVNULL
from sys import executable

from requests import Response, get
from packaging import version

# define constants for CLI colors:
RED = "\033[91m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
BLUE = "\033[94m"
RESET = "\033[0m"


def get_pypi_versions() -> Response:
    """
    Get the versions of qualysdk available on PyPI.

    Returns:
        Response: The response object from the request.
    """

    return get("https://pypi.org/pypi/qualysdk/json")


def parse_pypi_response(response: Response) -> dict:
    """
    Parse the response from PyPI.

    Args:
        response (Response): The response object from the request.

    Returns:
        dict: The parsed response.

    Raises:
        HTTPError: If the response is not successful
    """

    response.raise_for_status()
    return delete_yanked_versions(response.json())


def delete_yanked_versions(pypi_data: dict) -> dict:
    """
    Removes yanked versions from the versions dictionary.

    Helper function for parse_pypi_response.

    Args:
        pypi_data (dict): The versions dictionary.

    Returns:
        dict: The versions dictionary with yanked versions removed.
    """

    return {k: v for k, v in pypi_data.items() if not (k == "yanked" and v)}


def show_update_info(pypi_data: dict) -> None:
    """
    Print the update information to the console.

    Args:
        pypi_data (dict): The parsed response from PyPI.
    """

    vsn = pypi_data["info"].get("version")
    release_data = pypi_data["releases"].get(vsn)[0]
    digests = release_data.get("digests")
    if not digests:
        digests = {"blake2_256": "N/A", "sha256": "N/A", "md5_digest": "N/A"}

    print(f"🟧 {YELLOW}An update is available!{RESET}")
    print(f"📅 Latest Version: {YELLOW}v{vsn}{RESET}")
    print(f"📅 Release Date: {YELLOW}{release_data.get('upload_time')}{RESET}")
    print(f"✅ MD5: {RED}{digests.get('md5')}{RESET}")
    print(f"✅ BLAKE2b_256: {RED}{digests.get('blake2b_256')}{RESET}")
    print(f"✅ SHA256: {RED}{digests.get('sha256')}{RESET}")
    print(
        f"🐈 {GREEN}GitHub Release Notes:{RESET}{BLUE} https://github.com/0x41424142/qualysdk/releases/tag/v{vsn} {RESET}"
    )
    print(
        f"🐍 {GREEN}PyPI Page:{RESET}{BLUE} https://pypi.org/project/qualysdk/{vsn}/ {RESET}"
    )


def check_installed_version() -> version.Version:
    """
    Use subprocess to get the installed version of qualysdk
    for the given environment.

    Returns:
        version.Version: The installed version of qualysdk.
    """

    return version.parse(
        run(
            [executable, "-m", "pip", "show", "qualysdk"],
            capture_output=True,
            text=True,
        )
        .stdout.split("\n")[1]
        .split(": ")[1]
    )


def prompt_for_install() -> bool:
    """
    Prompt the user to install the latest version of qualysdk.

    Returns:
        bool: True if the user wants to install, False if not.
    """

    response = input(
        f"Would you like to install the {GREEN}latest version{RESET} over current version: {YELLOW}v{check_installed_version()}{RESET}? [y/N]: "
    )
    return response.lower() in ["y", "yes"]


def main() -> int:
    """
    Entry point for the updater CLI tool.

    Returns:
        int: 0 if successful, 1 if not.
    """

    parser = ArgumentParser(description="Check for and install updates for qualysdk")

    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Display the current version of qualysdk",
    )

    parser.add_argument(
        "-c",
        "--check",
        action="store_true",
        help="Check if a new version of qualysdk is available",
    )

    parser.add_argument(
        "-i",
        "--install",
        action="store_true",
        help="Install the latest version of qualysdk",
    )

    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Bypass the confirmation prompt when installing",
        default=False,
    )

    args = parser.parse_args()

    # If every arg is False or only -y was specified,
    #  print the help message and exit.
    if not any(vars(args).values()) or (args.yes and not args.install):
        parser.print_help()
        return 1

    if args.version:
        print(
            f"{YELLOW}Qualysdk version currently installed: v{check_installed_version()}{RESET}"
        )
        return 0

    if args.check or args.install:
        response = get_pypi_versions()
        # yanked versions are automatically removed
        pypi_data = parse_pypi_response(response)

        latest_version = version.parse(pypi_data.get("info").get("version"))
        current_version = check_installed_version()
        if latest_version > current_version:
            show_update_info(pypi_data)
            if args.install:
                if args.yes or prompt_for_install():
                    print(f"🟧 {YELLOW}Installing v{latest_version}...{RESET}")
                    try:
                        check_call(
                            [
                                executable,
                                "-m",
                                "pip",
                                "install",
                                "--upgrade",
                                "qualysdk",
                            ],
                            stdout=DEVNULL,
                        )
                        print(
                            f"✅ {GREEN}Qualysdk has been updated to v{latest_version}{RESET}"
                        )
                    except Exception as e:
                        print(f"{RED}Qualysdk update failed: {type(e).__name__}{RESET}")
                        return 1
                    return 0
                else:
                    print(f"{RED}Upgrade cancelled.{RESET}")
                    return 0
            return 0
        else:
            print(f"✅ {GREEN}Qualysdk is up to date (v{current_version}).{RESET}")
            return 0


if __name__ == "__main__":
    main()



"""
CLI script to quickly perform Global AssetView
(GAV) operations using qualysdk.
"""

from argparse import ArgumentParser, Namespace
from typing import Union
from json import dump

from qualysdk import TokenAuth, write_excel, BaseList
from qualysdk.gav import *


def parse_asset_id(value):
    try:
        return int(value)
    except ValueError:
        return value


def write_json(data: dict, output: str) -> None:
    with open(output, "w") as f:
        dump(data, f, indent=2)
    print(f"Data written to {output}.")


def cli_findings(auth: TokenAuth, args: Namespace, endpoint: str) -> None:
    kwargs = dict(args.kwarg) if args.kwarg else {}
    if "page_count" in kwargs:
        kwargs["page_count"] = int(kwargs["page_count"])
    for kwarg in kwargs:
        if str(kwargs[kwarg]).lower() == "true":
            kwargs[kwarg] = True
        elif str(kwargs[kwarg]).lower() == "false":
            kwargs[kwarg] = False

    match endpoint:
        case "count_assets":
            result = count_assets(auth, **kwargs)
        case "get_all_assets":
            result = get_all_assets(auth, **kwargs)
        case "get_asset":
            kwargs["assetId"] = args.asset_id
            result = get_asset(auth, **kwargs)
        case "query_assets":
            result = query_assets(auth, **kwargs)
        case _:
            raise ValueError(f"Invalid endpoint: {endpoint}.")

    # If the result object does NOT have the len() method available,
    # we need to wrap it in a BaseList:
    if not hasattr(result, "__len__") and endpoint != "count_assets":
        bl = BaseList()
        bl.append(result)
        result = bl

    (
        write_excel(result, args.output)
        if endpoint != "count_assets"
        else write_json(result, args.output)
    )


def main():
    parser = ArgumentParser(
        description="CLI script to quickly perform Global AssetView (GAV) operations using qualysdk"
    )
    parser.add_argument(
        "-u", "--username", required=True, help="Qualys username", type=str
    )
    parser.add_argument(
        "-p", "--password", required=True, help="Qualys password", type=str
    )
    parser.add_argument(
        "-P",
        "--platform",
        help="Qualys platform",
        default="qg3",
        choices=["qg1", "qg2", "qg3", "qg4"],
    )

    # subparser for action:
    subparsers = parser.add_subparsers(dest="action", help="Action to perform")

    # get_findings action:
    count_assets_parser = subparsers.add_parser(
        "count_assets",
        help="Count how many assets match a GAV QQL filter & save to a JSON file.",
    )
    count_assets_parser.add_argument(
        "-o",
        "--output",
        help="Output JSON file to write results to",
        type=str,
        default="gav_asset_count.json",
    )
    count_assets_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    get_all_assets_parser = subparsers.add_parser(
        "get_all_assets", help="Pull a list of all assets in the GAV API."
    )

    get_all_assets_parser.add_argument(
        "-o",
        "--output",
        help="Output xlsx file to write results to",
        type=str,
        default="gav_all_assets.xlsx",
    )

    get_all_assets_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    get_asset_parser = subparsers.add_parser(
        "get_asset",
        help="Get a specific asset by asset ID (not host ID) & save to a JSON file.",
    )

    get_asset_parser.add_argument(
        "asset_id",
        help="The ID of the asset to get.",
        type=parse_asset_id,
    )

    get_asset_parser.add_argument(
        "-o",
        "--output",
        help="Output xlsx file to write results to",
        type=str,
        default="gav_asset.xlsx",
    )

    get_asset_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    query_assets_parser = subparsers.add_parser(
        "query_assets", help="Query assets based on a GAV QQL filter."
    )

    query_assets_parser.add_argument(
        "-o",
        "--output",
        help="Output xlsx file to write results to",
        type=str,
        default="gav_queried_assets.xlsx",
    )

    query_assets_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    args = parser.parse_args()

    # create TokenAuth object
    auth = TokenAuth(args.username, args.password, platform=args.platform)

    # perform action
    if args.action == "count_assets":
        cli_findings(auth=auth, args=args, endpoint="count_assets")
    elif args.action == "get_all_assets":
        cli_findings(auth=auth, args=args, endpoint="get_all_assets")
    elif args.action == "get_asset":
        cli_findings(auth=auth, args=args, endpoint="get_asset")
    elif args.action == "query_assets":
        cli_findings(auth=auth, args=args, endpoint="query_assets")
    else:
        parser.print_help()
        exit(1)


if __name__ == "__main__":
    main()



"""
user.py - Contains the User data class.
"""

from typing import Union
from datetime import datetime
from dataclasses import dataclass, field, asdict

from ...base.base_list import BaseList


@dataclass
class User:
    """
    A user login in Qualys.
    """

    USER_LOGIN: str = field(metadata={"description": "User login name."}, default=None)

    USER_ID: Union[str, int] = field(metadata={"description": "User ID."}, default=None)

    EXTERNAL_ID: str = field(metadata={"description": "External ID."}, default=None)

    CONTACT_INFO: dict = field(
        metadata={"description": "Contact information."}, default=None
    )

    # BELOW ARE PARSED OUT OF CONTACT_INFO

    FIRSTNAME: str = field(metadata={"description": "First name."}, default=None)

    LASTNAME: str = field(metadata={"description": "Last name."}, default=None)

    TITLE: str = field(metadata={"description": "Title."}, default=None)

    PHONE: str = field(metadata={"description": "Phone number."}, default=None)

    FAX: str = field(metadata={"description": "Fax number."}, default=None)

    EMAIL: str = field(metadata={"description": "Email address."}, default=None)

    COMPANY: str = field(metadata={"description": "Company name."}, default=None)

    ADDRESS1: str = field(metadata={"description": "Address line 1."}, default=None)

    ADDRESS2: str = field(metadata={"description": "Address line 2."}, default=None)

    CITY: str = field(metadata={"description": "City."}, default=None)

    COUNTRY: str = field(metadata={"description": "Country."}, default=None)

    STATE: str = field(metadata={"description": "State."}, default=None)

    ZIP_CODE: str = field(metadata={"description": "Zip code."}, default=None)

    TIME_ZONE_CODE: str = field(
        metadata={"description": "Time zone code."}, default=None
    )

    # END PARSED OUT OF CONTACT_INFO

    USER_STATUS: str = field(metadata={"description": "User status."}, default=None)

    CREATION_DATE: Union[str, datetime] = field(
        metadata={"description": "Creation date."}, default=None
    )

    LAST_LOGIN_DATE: Union[str, datetime] = field(
        metadata={"description": "Last login date."}, default=None
    )

    USER_ROLE: str = field(metadata={"description": "User role."}, default=None)

    BUSINESS_UNIT: str = field(metadata={"description": "Business unit."}, default=None)

    UNIT_MANAGER_POC: int = field(
        metadata={"description": "Unit manager point of contact."}, default=None
    )

    MANAGER_POC: int = field(
        metadata={"description": "Manager point of contact."}, default=None
    )

    UI_INTERFACE_STYLE: str = field(
        metadata={"description": "UI interface style."}, default=None
    )

    PERMISSIONS: dict = field(metadata={"description": "Permissions."}, default=None)

    # BELOW ARE PARSED OUT OF PERMISSIONS

    CREATE_OPTION_PROFILES: bool = field(
        metadata={"description": "Create option profiles."}, default=None
    )

    PURGE_INFO: bool = field(
        metadata={"description": "Purge information."}, default=None
    )

    ADD_ASSETS: bool = field(metadata={"description": "Add assets."}, default=None)

    EDIT_REMEDIATION_POLICY: bool = field(
        metadata={"description": "Edit remediation policy."}, default=None
    )

    EDIT_AUTH_RECORDS: bool = field(
        metadata={"description": "Edit authentication records."}, default=None
    )

    # END PARSED OUT OF PERMISSIONS

    NOTIFICATIONS: dict = field(
        metadata={"description": "Notifications."}, default=None
    )

    # BELOW ARE PARSED OUT OF NOTIFICATIONS
    LATEST_VULN: str = field(
        metadata={"description": "Latest vulnerability."}, default=None
    )

    MAP: str = field(metadata={"description": "Map."}, default=None)

    SCAN: str = field(metadata={"description": "Scan."}, default=None)

    DAILY_TICKETS: bool = field(
        metadata={"description": "Daily tickets."}, default=None
    )

    # END PARSED OUT OF NOTIFICATIONS

    ASSIGNED_ASSET_GROUPS: dict = field(
        metadata={"description": "Assigned asset groups."}, default=None
    )

    # BELOW ARE PARSED OUT OF ASSIGNED_ASSET_GROUPS

    ASSET_GROUP_TITLE: BaseList[str] = field(
        metadata={"description": "Asset group title."}, default=None
    )  # May need to be a list in the future?

    # END PARSED OUT OF ASSIGNED_ASSET_GROUPS

    def __post_init__(self):
        DT_FIELDS = ["CREATION_DATE", "LAST_LOGIN_DATE"]
        INT_FIELDS = ["USER_ID", "UNIT_MANAGER_POC", "MANAGER_POC"]
        BOOL_FIELDS = [
            "CREATE_OPTION_PROFILES",
            "PURGE_INFO",
            "ADD_ASSETS",
            "EDIT_REMEDIATION_POLICY",
            "EDIT_AUTH_RECORDS",
            "DAILY_TICKETS",
        ]

        # Parse out the nested fields

        if self.CONTACT_INFO:
            for key, value in self.CONTACT_INFO.items():
                setattr(self, key, value if value else None)

        if self.PERMISSIONS:
            for key, value in self.PERMISSIONS.items():
                setattr(self, key, value if value else None)

        if self.NOTIFICATIONS:
            for key, value in self.NOTIFICATIONS.items():
                setattr(self, key, value if value else None)

        if self.ASSIGNED_ASSET_GROUPS:
            for key, value in self.ASSIGNED_ASSET_GROUPS.items():
                setattr(self, key, value if value else None)

        # Convert the fields to the correct types

        for field in DT_FIELDS:
            if getattr(self, field):
                if getattr(self, field) == "N/A":  # Some fields have N/A as a value
                    setattr(self, field, None)
                else:
                    setattr(
                        self,
                        field,
                        datetime.strptime(getattr(self, field), "%Y-%m-%dT%H:%M:%S%z"),
                    )

        for field in INT_FIELDS:
            if getattr(self, field):
                setattr(self, field, int(getattr(self, field)))

        for field in BOOL_FIELDS:
            if getattr(self, field):
                setattr(self, field, bool(getattr(self, field)))

        if self.ASSET_GROUP_TITLE:
            bl = BaseList()
            data = self.ASSET_GROUP_TITLE
            if not isinstance(data, list):
                data = [data]
            for item in data:
                bl.append(item)
            self.ASSET_GROUP_TITLE = bl

        if self.MAP == "none":
            self.MAP = None

        if self.SCAN == "none":
            self.SCAN = None

    def __dict__(self):
        return asdict(self)

    def __str__(self):
        return self.USER_LOGIN

    def keys(self):
        return self.__dict__().keys()

    def values(self):
        return self.__dict__().values()

    def items(self):
        return self.__dict__().items()

    def valid_values(self):
        return {k: v for k, v in self.__dict__().items() if v}

    def to_dict(self):
        return self.__dict__()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)
